
// SOURCE CODE HEADER
//
// ecdl2K-108.32bit.c (1.1.0)
// Purpose: Fast arithmetic for computing discrete logs on elliptic curves.
// Copyright: Robert J. Harley, 1997-1999.
// Contact: Robert.Harley@inria.fr
// Legalese: This source code is subject to the GNU General Public License v2
// Modified for Windows NT by B. R. Gladman (brian.gladman@btinternet.com) 
// with help from Jeff Gilchrist (jeff.gilchrist@entrust.com).

// README - QUICKSTART
// 
// 1. Sign up your participation at: http://cristal.inria.fr/bin/ecdldb.
// 
// 2. Create a new directory and put the files 'ecdl.exe' and ecdl.ini' 
//    into it. Other files generated by the program will be placed in
//    this directory. The file 'ecdl.log' records a log of the actions
//    taken by the program. This can be deleted periodically if you wish. 
//    The file (or files) 'state.x.bin', where x is a digit, record the 
//    program's progress and should not be interfered with.
// 
// 3. Edit the file 'ecdl.ini' to include your own details as described
//    within the file.
// 
// 4. Run ecdl using the run command with the command line 'ecdl -i' to
//    install the 'ECDL2K-108 Service Client' (it can be removed using 
//    'ecdl -u'). Alternatively issue these commands using a command
//    prompt.
// 
// 5. If you are running Windows NT, select 'Services' in the Windows 
//    Control Panel, find and select the 'ECDL2K-108 Service Client' 
//    and start it using the start button (it will start automatically 
//    on the next machine restart if you do not do this).  If you are 
//    running Windows 95/98, you will need to reboot the computer to 
//    start the ecdl service running in the background. 
// 
// 6. The program will place details of the points it finds in a file 
//    named 'dist.points'.  If running in batch mode, rename this file
//    every few days to dist.points.nov.25 (but with the current date 
//    at the end rather than Nov.25) and email this renamed file to the 
//    following address: 'ecdl2K-108@pauillac.inria.fr'
// 
// 7. New points will be put in a new dist.points file so step 6 can be
//    repeated every few days.
// 
// If you wish to run ecdl as a conventional program rather than a service, 
// you can do this using the command 'ecdl -r'. The program will then run 
// until terminated manually using Control-C (or by a computer restart). 
// Other commands are:
// 
//   'ecdl -h' :  provide a summary of ecdl commands
//   'ecdl -v' :  display version info and install status
// 
// This port of Robert Harley's ecdl2K-108 program has been implemented for
// use on either Windows 9x or Windows NT systems as a system service. It 
// will take advantage of MMX capabilities as well as multiprocessors (if 
// present) on Windows NT systems.  It only uses idle processor time and 
// should not interfere in any way with other activities running on the 
// same machines.
// 
// The port was implemented by Brian Gladman (brian.gladman@btinternet.com)
// with help from Jeff Gilchrist (jeff.gilchrist@entrust.com).

// VERSION INFORMATION
//
// Version 0.9.0: 31 September 1999
// - Does 66 k iters/sec with gcc 2.95+Linux+500 MHz Pentium III (PARAL=20).
// Version 0.99.0: 24 November 1999
// - Does 82 k iters/sec with gcc 2.95+Linux+500 MHz Pentium III (PARAL=32)
// and 145 k when using MMX!
// Version 1.0.0: 3 December 1999
// Version 1.1.0: 28th December - implemented as a Windows NT or 9x Service
//
// This is a conversion of the original source code by Dr. B. R. Gladman to
// take on board Windows9x and Windows NT systems. On a 600MHz PIII system
// it gives 160k and 128k iters/sec with and without MMX respectively.
//
// 1.1.1<brg,jg> released on 4th January 2000 - fix for bug when proxy is 
// used and removal of dependence on winsock 2 to help Windows 95 users. 
// Some minor performance improvements for WIndows NT.

#define VARIANT "WIN32"

// TESTING
//
// You can test the binary like this:
//
//      ./ecdl.exe test by me@here on bla
//
// The test output should match this sample except for the reported
// times and rates:
//
//		>./ecdl.exe test by me@here on bla
//		This is Rob's ECDL program for Unix32, version 1.0.0 MMX
//		Mode ...... test
//		Email ..... me@here
//		Machine ... bla
//		Generating 32 new starting points.
//		Computing iterations...
//
//		Test point found at: Fri Nov 26 18:16:41 1999
//		TEST-L0|Unix32|1.0.0 MMX|
//		TEST-L1|EF7DC6C12ACAF01D48CE4D44EB6|5163373649C868AD3265FACCF99|
//		TEST-L2|me@here|
//		TEST-L3|000000004887|
//		TEST-L4|bla|
//		Iterations used = 18567.
//		Total iterations = 594144.
//		Rate = 144913 per second.
//
//		Test point found at: Fri Nov 26 18:17:01 1999
//		TEST-L0|Unix32|1.0.0 MMX|
//		TEST-L1|B0152E85E4E22B6EE3C7120FC46|026063744AFC9DB003098294BD6|
//		TEST-L2|me@here|
//		TEST-L3|00000001BACA|
//		TEST-L4|bla|
//		Iterations used = 113354.
//		Total iterations = 3.62733e+06.
//		Rate = 147153 per second.
//	
//		Test point found at: Fri Nov 26 18:17:08 1999
//		TEST-L0|Unix32|1.0.0 MMX|
//		TEST-L1|607BB600DECAF410A7B62ED3DD3|F43D17AAF642BBF4EBFAA9C1984|
//		TEST-L2|me@here|
//		TEST-L3|000000023C69|
//		TEST-L4|bla|
//		Iterations used = 146537.
//		Total iterations = 4.68918e+06.
//		Rate = 146950 per second.
//	
//		[...and so on...]

#define WIN32_LEAN_AND_MEAN

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <windows.h>
#include <process.h>
#include <time.h>
#include <stdlib.h>
#include <winsock.h>

#include "myservice.h"
#include "osinfo.h"

// OPERATING MODES

typedef enum { Bad, Test, Batch, Mail, Alt, Http } ModeType;

// GLOBAL VARIABLES

int					gotSignal = 0;			// termination flag 
unsigned long 		thr_active = 0;			// global for thread state on each processor
CRITICAL_SECTION	crit;					// critical section semaphore
ModeType			mode;					// program mode
char				st_fpath[_MAX_PATH];	// state file path
char				dp_fpath[_MAX_PATH];	// state file path
char			   *argv3;					// globals for command line arguments
char			   *argv5;

// This source and the related files to implement ecdl as a service were
// provided by Jeff Gilchrist <Jeff.Gilchrist@entrust.com>

CMyService MyService;

int main(int argc, char* argv[])
{
    // Parse for standard arguments (install, uninstall, version, etc).
    if(!MyService.ParseStandardArgs(argc, argv)) 
	{
        // Didn't find any standard args so start the service.
        // Uncomment the DebugBreak line below to enter the debugger
        // when the service is started.
        // DebugBreak();
        MyService.StartService();
    }

    // When we get here, the service has been stopped.
    return MyService.m_Status.dwWin32ExitCode;
}

// Name of file for writing distinguished points to (as well as stdout)

#ifndef DIST_POINTS_FILENAME
#define DIST_POINTS_FILENAME "dist.points"
#endif

// Name of file for saving state after each distinguished point or when
// stopping due to a SIGINT or SIGTERM.  It is read when restarting

#ifndef SAVED_STATE_FILENAME
#  define SAVED_STATE_FILENAME "state.0.bin"
#endif

// The state is checked and saved every 2^SAVED_STATE_SHIFT iterations
// and also after a distinguished point is found

#ifndef SAVED_STATE_SHIFT
#  define SAVED_STATE_SHIFT (26)
#endif

#if SAVED_STATE_SHIFT < 20 || SAVED_STATE_SHIFT > 31
#  error Please define SAVED_STATE_SHIFT in the range 20..31.
#endif

// Number of iterations to run "in parallel".  This is done so that PARAL 
// inversions can be replaced by one inversion and 3*PARAL-3 mults
// Try something roughly in the range 20 to 50

#ifndef PARAL
#  define PARAL (32)
#endif

#if PARAL < 1
#  error Please define PARAL to be greater than 0.
#endif

// Define PROD to choose a different implementation of the critical
// GF2ProductNNxNN() and product() functions.  The default is 1
// Try 2, 3, 4 or 5.  They might be faster.  Then again maybe not!

#  ifndef PROD
#    define PROD 4
#  endif

#  if PROD < 1 || PROD > 5
#    error Please define PROD in the range 1..5.
#  endif

// TYPE AND BYTE ORDER INFORMATION
//
// 128 bit numeric data items in this source code are specified as 32 bit
// word sequences with the most significant words on the left, that is in 
// natural numeric (big endian) order.  If needed the macros data_p128() 
// and data_u128() can be defined to put all such data items into little 
// endian order if either of the 128 bit structures below (u128, poly128) 
// are assembled in memory in this form.  Ordering is also set by the macro 
// data_f(a,b,c,d) that is used used later in the code. 
//
// Since MMX instructions are for little endian pentium processors the input 
// and output memory blocks for MMX operations will be in this order and the 
// mmx64 type below MUST be set so that the two 32 bits words within it are 
// in this order.

// Since the default order for data and structures in this source code is 
// big endian {_3,_2,_1,_0}, there will normally be a change of order of 
// the two 32 bit words in a 64 bit block when an MMX 64 bit item is moved 
// into a 64 or 128 bit data structures (u64, u128 and poly128). However,
// if CHANGE_ORDER is defined, the structures are mapped in little endian 
// order and the big endian data items are inverted to account for this 
// using data_p128() and data_u128(). Visual C++ maps the declaration order 
// for items within structures into increasing addresses. The use of
// compiler that maps to decreasing addresses will require changes to the
// order conventions set out here.

#define CHANGE_ORDER

#ifdef CHANGE_ORDER

#define	data_p128(a,b,c,d)	{ d, c, b, a }
#define	data_u128(a,b,c,d)	{ d, c, b, a }

typedef unsigned int uint;	// Abbreviations used when bit-length does not matter

typedef unsigned char u8;
typedef unsigned int u32;

typedef struct { u32 _0, _1; } u64;
typedef struct { u32 _0, _1, _2, _3; } u128;

// Polynomials over the two-element field GF(2) = Z/2Z

typedef struct { u32 _0, _1, _2, _3; } poly128;

#else

#define	data_p128(a,b,c,d)	{ a, b, c, d }
#define	data_u128(a,b,c,d)	{ a, b, c, d }

typedef unsigned int uint;	// Abbreviations used when bit-length does not matter

typedef unsigned char u8;
typedef unsigned int u32;

typedef struct { u32 _1, _0; } u64;
typedef struct { u32 _3, _2, _1, _0; } u128;

// Polynomials over the two-element field GF(2) = Z/2Z

typedef struct { u32 _3, _2, _1, _0; } poly128;

#endif

// MMX TYPES - ENDIANNESS, ALIGNMENT AND PACKING MATTER HERE

// These structures MUST be assembled so that the least signaificant
// 32 bit word (_0) in each 64 bit structure is at the lower memory 
// address.  The two 32 bit words must be contiguous in memory and
// the structure must be aligned on a 64 bit (8 byte) boundary.

typedef struct 
{	u32 _0, _1;
} mmx64;

typedef struct { mmx64 _0, _1; } mmx128;			// 128 bit mmx data

// 128 bit data to mmx128 - we always want little endian here

#define data_m128(a,b,c,d)	{{ d, c }, { b, a }}

// These are used a few places for types poly128

static const poly128 p128_0 = data_p128( 0, 0, 0, 0);
static const poly128 p128_1 = data_p128( 0, 0, 0, 1);

// The number of cases used in pseudo-random function

#define CASES (7)

// Population count used to decide if a point is distinguished

// Expect a given popc to occur one time in 2^108/binomial(109, popc)
// if it's odd (even ones never occur)

// 1 time in 1422267 or so (for testing)

#define TEST_POP_COUNT (29)

// 1 time in 1407700775 or so (for real)

#define DIST_POP_COUNT (23)

// THE CHALLENGE
//
// Data defining which ECDL problem to solve
//
// Field is GF(2^109) represented as (Z/2Z)[t] / (t^109+t^9+t^2+t+1)
// (t^109+t^9+t^2+t+1 is irreducible in (Z/2Z)[t])
//
// Curve from Certicom ECC2K-108 problem is y^2 + x*y = x^3 + x^2 + 1
// over GF(2^109).
//
// Group order is G = 2^109+1-w^109-conj(w)^109 where w = (1-sqrt(-7))/2.
// Use subgroup of order G/2 = 324518553658426701487448656461467 (prime)
//
// Frobenius is multiplication by w i.e., by
// 138423345589698157369693034392981 modulo g

static const poly128	p128_a = data_p128(0, 0, 0, 1);
static const u128		order  = data_u128(0x0FFF, 0xFFFFFFFF, 0xFFA621B0, 0x2C383E9B);

// Points from Certicom ECC2K-108 problem: 

#define XP      data_u128(0x0478, 0xC46CC963, 0x38CED915, 0x65E17257)
#define YP      data_u128(0x1E79, 0x65E4A3AF, 0xB73A48FC, 0x9AB790E9)
#define XQ      data_u128(0x1FF0, 0xCE5EC618, 0x93F2119C, 0x3077C59E)
#define YQ      data_u128(0x1F20, 0xE9B010AC, 0x691C9B87, 0xB438241D)

static const poly128 xP = XP, xQ = XQ, yP = YP, yQ = YQ;

static unsigned short	init = 0;
static unsigned long	w = 521288629, z = 362436069;

static unsigned long lrand(void)
{	unsigned long  t;

	if(!init)
	{
		init = 1; cycles(&t); w ^= t; cycles(&t); z ^= t; 
	}

	z = 36969 * (z & 65535) + (z >> 16);
	w = 18000 * (w & 65535) + (w >> 16);

	return (z << 16) + w;
}

// ARITHMETIC ON 128 BIT VARIABLES

// Return whether x >= y, for 0 <= x, y < 2^128

static inline int ge(const u128 *x, const u128 *y) 
{
	return x->_3 > y->_3 || x->_3 == y->_3  && (x->_2 > y->_2 
		|| x->_2 == y->_2 && (x->_1 > y->_1 || x->_1 == y->_1 && x->_0 >= y->_0));
} 

// Return x-y for 0 <= y <= x < 2^128

static inline void diff(u128 *x, const u128 *y) 
{	u32 c, d;

	c = x->_0 < y->_0; x->_0 -= y->_0;
	d = x->_1 < c; x->_1 -= c; c = d | (x->_1 < y->_1); x->_1 -= y->_1;
	d = x->_2 < c; x->_2 -= c; c = d | (x->_2 < y->_2); x->_2 -= y->_2;
	x->_3 -= c; x->_3 -= y->_3;
}

// Return sum x+y modulo m, 0 <= x, y < _1 <= 2^127

static inline void sumMod(u128 *x, const u128 *y, const u128 *m) 
{	u32 c;

	x->_0 += y->_0; c = x->_0 < y->_0;

	x->_1 += c; c = x->_1 < c; x->_1 += y->_1; c |= x->_1 < y->_1;

	x->_2 += c; c = x->_2 < c; x->_2 += y->_2; c |= x->_2 < y->_2;
	
	x->_3 += c; x->_3 += y->_3;

	if(ge(x, m)) 
		
		diff(x, m);
}

// Return double x << 1 modulo m, where 0 <= x < m <= 2^127

static inline void doubleMod(u128 *x, const u128 *m) 
{
	x->_3 <<= 1; x->_3 |= x->_2 >> 31;
	x->_2 <<= 1; x->_2 |= x->_1 >> 31;
	x->_1 <<= 1; x->_1 |= x->_0 >> 31;
	x->_0 <<= 1;

	if(ge(x, m)) 
	  
		diff(x, m);
}

// Increment x modulo m, for 0 <= x < m < 2^128

static inline void incMod(u128 *x, const u128 *m) 
{
	if(!++x->_0) if(!++x->_1) if(!++x->_2) ++x->_3;

	if(x->_0 == m->_0 && x->_1 == m->_1 && x->_2 == m->_2 && x->_3 == m->_3) 
	{
		x->_0 = x->_1 = x->_2 = x->_3 = 0;
	}
}

// FUNCTIONS FOR POLYNOMIAL OPERATIONS

// Return whether x == y, for polys over Z/2Z

static inline int equal(const poly128 *x, const poly128 *y) 
{
	return x->_0 == y->_0 && x->_1 == y->_1 && x->_2 == y->_2 && x->_3 == y->_3;
}

// Return x XOR y, in other words add them as polys over Z/2Z

static inline void xor(poly128 *x, const poly128 *y) 
{
	x->_0 ^= y->_0; x->_1 ^= y->_1; x->_2 ^= y->_2; x->_3 ^= y->_3;
}

// Returns highest set bit of x (in position), or 0 if x = 0

static inline u32 top_bit(u32 x) 
{
	x |= x >> 1; x |= x >> 2; x |= x >> 4; x |= x >> 8; x |= x >> 16;

	return x ^ (x >> 1);
} 

// Stuff for checking that a point is distinguished

// Counts bits set in a 109-bit poly over Z/2Z

static uint popc(const poly128 *n) 
{	u32			n1, n2, nt;
	const u32	mask1 = 0x55555555, mask2 = 0x33333333, mask3 = 0x0F0F0F0F;

	//	Full adder

	n1 = n->_0 ^ n->_1; n2 = n->_0 & n->_1; n2 |= n1 & n->_2; n1 ^= n->_2;

	// Partial popc
  
	n2 -= n2 >> 1 & mask1;						// Count 2-by-2
	n2 = (n2 & mask2) + ((n2 >> 2) & mask2);	// Count 4-by-4
	n2 += n2 >> 4; n2 &= mask3;			        // Count 8-by-8

	// Partial popc

	n1 -= (n1 >> 1) & mask1; n1 = (n1 & mask2) + ((n1 >> 2) & mask2); 
	n1 += n1 >> 4; n1 &= mask3; nt = n->_3;

	// Partial popc

	nt -= nt >> 1 & mask1; nt = (nt & mask2) + ((nt >> 2) & mask2); 
	nt += nt>> 4; nt &= mask3;

	// Accumulate pieces

	n1 += (n2 << 1) + nt; n1 += n1 >> 8;

	return (n1 + (n1 >> 16)) & 255;
}

// FINITE FIELD ARITHMETIC

// Square x, n times, in the field GF(2^109) i.e., as a poly in (Z/2Z)[t]
// reduced modulo t^109+t^9+t^2+t+1.  Degree x < 109.

// Some tables in the squaring operation in Robert's code are now generated 
// dynamically using the following macros

static u32 tt[4] = { 0, 1, 4, 5 };

#define tgen_2(i)	4096 * tt[(i >> 6) & 3] + 256 * tt[(i >> 4) & 3] \
						+ 16 * tt[(i >> 2) & 3] + tt[i & 3];

#define tgen_4(i)	(i & 128) << 21 | (i & 64) << 18 | (i & 32) << 15 | (i & 16) << 12 \
					 | (i & 8) << 9 | (i & 4) << 6 | (i & 2) << 3 | (i & 1) 

// this table is used in changing base and is always stored 
// in MMX order

static const mmx128 matrix[218] =
{   data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1fff, 0xffffffff, 0xffffffff, 0xffffffff),
    data_m128(0x1fff, 0xffffffff, 0xffffffff, 0xfffffffe),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000001),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1fff, 0xffffffff, 0xffffffff, 0xfffffffd),
    data_m128(0x09ea, 0x1a34beec, 0x0a908559, 0xbdeda6ab),
    data_m128(0x1615, 0xe5cb4113, 0xf56f7aa6, 0x42125956),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1fff, 0xffffffff, 0xffffffff, 0xfffffffb),
    data_m128(0x0536, 0x810828a6, 0x03436150, 0x2fc46fd6),
    data_m128(0x1ac9, 0x7ef7d759, 0xfcbc9eaf, 0xd03b902d),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x13d4, 0x34697dd8, 0x15210ab3, 0x7bdb4d56),
    data_m128(0x1748, 0xb0c11920, 0x5dc6e0e9, 0x09a60ac2),
    data_m128(0x049c, 0x84a864f8, 0x48e7ea5a, 0x727d4794),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1fff, 0xffffffff, 0xffffffff, 0xfffffff7),
    data_m128(0x1e31, 0xe07b050a, 0x23a45550, 0x30fad343),
    data_m128(0x01ce, 0x1f84faf5, 0xdc5baaaf, 0xcf052cb4),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0a6d, 0x0210514c, 0x0686c2a0, 0x5f88dfac),
    data_m128(0x02c7, 0x94f5e66a, 0xaa1ee736, 0x6f735e1b),
    data_m128(0x08aa, 0x96e5b726, 0xac982596, 0x30fb81b7),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x07a8, 0x68d2fbb0, 0x2a421566, 0xf7b69aad),
    data_m128(0x0346, 0x9e40b570, 0xce15c900, 0x71b8ccc4),
    data_m128(0x04ee, 0xf6924ec0, 0xe457dc66, 0x860e5669),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0e91, 0x61823240, 0xbb8dc1d2, 0x134c1585),
    data_m128(0x0a5e, 0x6cc8e146, 0x486897c2, 0x0acd96e5),
    data_m128(0x04cf, 0x0d4ad306, 0xf3e55610, 0x19818360),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1fff, 0xffffffff, 0xffffffff, 0xffffffef),
    data_m128(0x1de5, 0xdf8084bb, 0x9970280c, 0xa3bf99c1),
    data_m128(0x021a, 0x207f7b44, 0x668fd7f3, 0x5c40662e),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1c63, 0xc0f60a14, 0x4748aaa0, 0x61f5a687),
    data_m128(0x1f55, 0x01b79a31, 0xd77e3e4d, 0x9fbddcc1),
    data_m128(0x0336, 0xc1419025, 0x903694ed, 0xfe487a46),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x14da, 0x0420a298, 0x0d0d8540, 0xbf11bf58),
    data_m128(0x0952, 0x396ca3d7, 0x5297c756, 0xf9cbb7ec),
    data_m128(0x1d88, 0x3d4c014f, 0x5f9a4216, 0x46da08b4),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x058f, 0x29ebccd5, 0x543dce6c, 0xdee6bc36),
    data_m128(0x1bfa, 0x18aa70fe, 0x45b2d97a, 0xcd057310),
    data_m128(0x1e75, 0x3141bc2b, 0x118f1716, 0x13e3cf26),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0f50, 0xd1a5f760, 0x54842acd, 0xef6d355a),
    data_m128(0x192a, 0xafd26253, 0x0fd989fc, 0x56604c4e),
    data_m128(0x167a, 0x7e779533, 0x5b5da331, 0xb90d7914),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x068d, 0x3c816ae1, 0x9c2b9200, 0xe3719988),
    data_m128(0x0095, 0xeac2d27f, 0xad933bef, 0xb6d89e6c),
    data_m128(0x0618, 0xd643b89e, 0x31b8a9ef, 0x55a907e4),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1d22, 0xc3046481, 0x771b83a4, 0x26982b0a),
    data_m128(0x0a5e, 0x542e870a, 0x9f98ad6e, 0x4e573666),
    data_m128(0x177c, 0x972ae38b, 0xe8832eca, 0x68cf1d6c),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x14bc, 0xd991c28c, 0x90d12f84, 0x159b2dca),
    data_m128(0x09c5, 0x27449b58, 0x9e98dabf, 0x2e783164),
    data_m128(0x1d79, 0xfed559d4, 0x0e49f53b, 0x3be31cae),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1fff, 0xffffffff, 0xffffffff, 0xffffffdf),
    data_m128(0x1c9f, 0xe648868c, 0xb844216d, 0x22fdfd76),
    data_m128(0x0360, 0x19b77973, 0x47bbde92, 0xdd0202a9),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1bcb, 0xbf010977, 0x32e05019, 0x477f3383),
    data_m128(0x00c4, 0x4daca9c6, 0xf0ce03ab, 0x2356535f),
    data_m128(0x1b0f, 0xf2ada0b1, 0xc22e53b2, 0x642960dc),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x18c7, 0x81ec1428, 0x8e915540, 0xc3eb4d0f),
    data_m128(0x1003, 0x982774e6, 0xec619f31, 0x52dd5203),
    data_m128(0x08c4, 0x19cb60ce, 0x62f0ca71, 0x91361f0c),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1eaa, 0x036f3463, 0xaefc7c9b, 0x3f7bb983),
    data_m128(0x0be2, 0xda6a67e3, 0x9960fc56, 0x32adbef9),
    data_m128(0x1548, 0xd9055380, 0x379c80cd, 0x0dd6077a),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x09b4, 0x08414530, 0x1a1b0a81, 0x7e237eb1),
    data_m128(0x1cf9, 0x11fd1f88, 0x9bf5df89, 0xd921e2f2),
    data_m128(0x154d, 0x19bc5ab8, 0x81eed508, 0xa7029c43),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x12a4, 0x72d947ae, 0xa52f8ead, 0xf3976fd8),
    data_m128(0x1ce4, 0x8176618a, 0x5063cf7b, 0x362cfc89),
    data_m128(0x0e40, 0xf3af2624, 0xf54c41d6, 0xc5bb9351),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0b1e, 0x53d799aa, 0xa87b9cd9, 0xbdcd786c),
    data_m128(0x0048, 0xa61ffc6d, 0x7bbe85cf, 0xb0ec51fa),
    data_m128(0x0b56, 0xf5c865c7, 0xd3c51916, 0x0d212996),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x17f4, 0x3154e1fc, 0x8b65b2f5, 0x9a0ae621),
    data_m128(0x19e3, 0xb34684b4, 0xae959235, 0x14e90a5d),
    data_m128(0x0e17, 0x82126548, 0x25f020c0, 0x8ee3ec7c),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1ea1, 0xa34beec0, 0xa908559b, 0xdeda6ab4),
    data_m128(0x128c, 0xffac3cf8, 0x8c2931be, 0x4db5cd19),
    data_m128(0x0c2d, 0x5ce7d238, 0x25216425, 0x936fa7ad),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1255, 0x5fa4c4a6, 0x1fb313f8, 0xacc0989d),
    data_m128(0x0eef, 0xf28924aa, 0x2f887351, 0x1c93544f),
    data_m128(0x1cba, 0xad2de00c, 0x303b60a9, 0xb053ccd2),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0d1a, 0x7902d5c3, 0x38572401, 0xc6e33310),
    data_m128(0x0c5c, 0x19ad60b0, 0x5e11225d, 0x09cbc60f),
    data_m128(0x0146, 0x60afb573, 0x6646065c, 0xcf28f51f),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x012b, 0xd585a4ff, 0x5b2677df, 0x6db13cd8),
    data_m128(0x01c3, 0x8c1277d0, 0x060b23fc, 0xf132bc82),
    data_m128(0x00e8, 0x5997d32f, 0x5d2d5423, 0x9c83805a),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1a45, 0x8608c902, 0xee370748, 0x4d305615),
    data_m128(0x179d, 0xb2113f0f, 0x7814a56c, 0x277a9ee3),
    data_m128(0x0dd8, 0x3419f60d, 0x9623a224, 0x6a4ac8f6),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x14bc, 0xa85d0e15, 0x3f315adc, 0x9cae6ccc),
    data_m128(0x157c, 0x5b345c1c, 0x643a6d33, 0x30fa7a39),
    data_m128(0x01c0, 0xf3695209, 0x5b0b37ef, 0xac5416f5),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0979, 0xb3238519, 0x21a25f08, 0x2b365b95),
    data_m128(0x0103, 0x5d70d551, 0xb4e6a8be, 0x5b6ff9c9),
    data_m128(0x087a, 0xee535048, 0x9544f7b6, 0x7059a25c),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x138a, 0x4e8936b1, 0x3d31b57e, 0x5cf062c8),
    data_m128(0x1c2f, 0x11203f3f, 0xcab69770, 0xdf433d07),
    data_m128(0x0fa5, 0x5fa9098e, 0xf787220e, 0x83b35fcf),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1fff, 0xffffffff, 0xffffffff, 0xffffffbf),
    data_m128(0x01cd, 0xe8de6eb8, 0x69e4ffe2, 0x7fd9aa80),
    data_m128(0x1e32, 0x17219147, 0x961b001d, 0x8026553f),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x193f, 0xcc910d19, 0x708842da, 0x45fbfaed),
    data_m128(0x127c, 0xe34869dd, 0x0640b805, 0x225b3c14),
    data_m128(0x0b43, 0x2fd964c4, 0x76c8fadf, 0x67a0c6f9),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1797, 0x7e0212ee, 0x65c0a032, 0x8efe6707),
    data_m128(0x14c6, 0xd4ca755e, 0x9f969267, 0x2567c9a5),
    data_m128(0x0351, 0xaac867b0, 0xfa563255, 0xab99aea2),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0188, 0x9b59538d, 0xe19c0756, 0x46aca6be),
    data_m128(0x0a9a, 0x331e30c9, 0xefa2fe39, 0x5aa1f939),
    data_m128(0x0b12, 0xa8476344, 0x0e3ef96f, 0x1c0d5f87),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x118f, 0x03d82851, 0x1d22aa81, 0x87d69a1f),
    data_m128(0x095f, 0xb83b634d, 0x284c5684, 0xe280d9ac),
    data_m128(0x18d0, 0xbbe34b1c, 0x356efc05, 0x655643b3),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0007, 0x304ee9cd, 0xd8c33e62, 0xa5baa407),
    data_m128(0x0076, 0x5efb7e85, 0x388546d3, 0xd82d749f),
    data_m128(0x0071, 0x6eb59748, 0xe04678b1, 0x7d97d098),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1d54, 0x06de68c7, 0x5df8f936, 0x7ef77307),
    data_m128(0x07ba, 0x78c41eed, 0x02b46968, 0x29df40af),
    data_m128(0x1aee, 0x7e1a762a, 0x5f4c905e, 0x572833a8),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x17c5, 0xb4d4cfc7, 0x32c1f8ac, 0x655b7df2),
    data_m128(0x1d45, 0x7d6e6970, 0xc19185e5, 0xb0a43683),
    data_m128(0x0a80, 0xc9baa6b7, 0xf3507d49, 0xd5ff4b71),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1368, 0x10828a60, 0x34361502, 0xfc46fd62),
    data_m128(0x1f9f, 0x0759c297, 0x1f06e180, 0x64dc124e),
    data_m128(0x0cf7, 0x17db48f7, 0x2b30f482, 0x989aef2c),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x19f2, 0x23fa3f11, 0x37ebbf13, 0xb243c5e5),
    data_m128(0x17b1, 0x46f4b00f, 0xa9cbe82c, 0xe76ab97d),
    data_m128(0x0e43, 0x650e8f1e, 0x9e20573f, 0x55297c98),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0548, 0xe5b28f5d, 0x4a5f1d5b, 0xe72edfb1),
    data_m128(0x0797, 0x4cc047a9, 0x517ec6f3, 0x4f38005c),
    data_m128(0x02df, 0xa972c8f4, 0x1b21dba8, 0xa816dfed),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x19c9, 0x02ecc314, 0xa0c79ef6, 0x6c59f913),
    data_m128(0x0544, 0xfba4f1d8, 0xe3f79daf, 0xdee1e43f),
    data_m128(0x1c8d, 0xf94832cc, 0x43300359, 0xb2b81d2c),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x163c, 0xa7af3355, 0x50f739b3, 0x7b9af0d8),
    data_m128(0x0408, 0xe42cecee, 0xd1b4c9d2, 0x2e18f5e8),
    data_m128(0x1234, 0x4383dfbb, 0x8143f061, 0x55820530),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0091, 0x4c3ff8da, 0xf77d0b9f, 0x61d8a3f4),
    data_m128(0x186e, 0x7aba658a, 0x7ff0c893, 0x16aa6668),
    data_m128(0x18ff, 0x36859d50, 0x888dc30c, 0x7772c59c),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0fe8, 0x62a9c3f9, 0x16cb65eb, 0x3415cc43),
    data_m128(0x1a79, 0xd9c93da7, 0x2891c56b, 0xc83cf966),
    data_m128(0x1591, 0xbb60fe5e, 0x3e5aa080, 0xfc293525),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x13c7, 0x668d0969, 0x5d2b246a, 0x29d214bb),
    data_m128(0x08a6, 0x97345aa9, 0xbb2dde90, 0x2816e691),
    data_m128(0x1b61, 0xf1b953c0, 0xe606fafa, 0x01c4f22a),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1d43, 0x4697dd81, 0x5210ab37, 0xbdb4d569),
    data_m128(0x146e, 0xe982be06, 0x27b2f0e0, 0x6a2904e5),
    data_m128(0x092d, 0xaf156387, 0x75a25bd7, 0xd79dd18c),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0519, 0xff5879f1, 0x1852637c, 0x9b6b9a33),
    data_m128(0x1cf7, 0x16f88427, 0xa0b5b4d1, 0x6100d523),
    data_m128(0x19ee, 0xe9a0fdd6, 0xb8e7d7ad, 0xfa6b4f10),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x04aa, 0xbf49894c, 0x3f6627f1, 0x5981313b),
    data_m128(0x0e63, 0xe6b11399, 0xc8c3ec99, 0xf86f51e9),
    data_m128(0x0ac9, 0x59f89ad5, 0xf7a5cb68, 0xa1ee60d2),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1ddf, 0xe5124954, 0x5f10e6a2, 0x3926a89e),
    data_m128(0x1a9c, 0x19bc4158, 0x11bbb1fb, 0xc1eae3f5),
    data_m128(0x0743, 0xfcae080c, 0x4eab5759, 0xf8cc4b6b),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x1a34, 0xf205ab86, 0x70ae4803, 0x8dc66620),
    data_m128(0x0553, 0x811dfa1a, 0xc53f7e1c, 0xf16331c3),
    data_m128(0x1f67, 0x7318519c, 0xb591361f, 0x7ca557e3),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x18b8, 0x335ac160, 0xbc2244ba, 0x13978c1e),
    data_m128(0x0e4f, 0xb2233cdc, 0x78ec2fd5, 0x9bff7ff6),
    data_m128(0x16f7, 0x8179fdbc, 0xc4ce6b6f, 0x8868f3e8),
    data_m128(0x0000, 0x00000000, 0x00000000, 0x00000000),
    data_m128(0x0257, 0xab0b49fe, 0xb64cefbe, 0xdb6279b0)
};

#define mm_xhi mm0
#define mm_xlo mm1
#define mm_yhi mm2
#define mm_ylo mm3

// Go from basis { t^k | 0 <= k < 109 } to basis { (t+1)^2^k | 0 <= k < 109 }
// (both over Z/2Z)

#define	core				\
	__asm	mov		eax,edx	\
	__asm	ror		edx,2	\
	__asm	and		eax,ecx	\
	__asm	pxor	mm_ylo,qword ptr [ebx + eax]mmx128._0		\
	__asm	pxor	mm_yhi,qword ptr [ebx + eax]mmx128._1		\
	__asm	mov		eax,edx	\
	__asm	ror		edx,2	\
	__asm	and		eax,ecx	\
	__asm	pxor	mm_ylo,qword ptr [ebx + eax + 64]mmx128._0	\
	__asm	pxor	mm_yhi,qword ptr [ebx + eax + 64]mmx128._1	\
	__asm	add		ebx,128

static void MMX_change_base(poly128 *res, const poly128 *x) 
{
	__asm
	{	pxor	mm_ylo,mm_ylo
		pxor	mm_yhi,mm_yhi
		mov		ebx,offset matrix
		mov		ecx,48

		mov		eax,x	// f_block(x->_0, 8, p);
		mov		edx,dword ptr[eax]poly128._0
		rol		edx,4
		core
		core
		core
		core
		core
		core
		core
		core

		mov		eax,x	// f_block(x->_1, 8, p + 56);
		mov		edx,dword ptr[eax]poly128._1
		rol		edx,4
		core
		core
		core
		core
		core
		core
		core
		core

		mov		eax,x	// f_block(x->_2, 8, p + 112);
		mov		edx,dword ptr[eax]poly128._2
		rol		edx,4
		core
		core
		core
		core
		core
		core
		core
		core

		mov		eax,x	// f_block(x->_3, 3, p + 168);
		mov		edx,dword ptr[eax]poly128._3
		rol		edx,4
		core
		core
		core
		
		mov		eax,edx
		and		eax,16
		pxor	mm_ylo,qword ptr [ebx + eax]mmx128._0
		pxor	mm_yhi,qword ptr [ebx + eax]mmx128._1

		// set result in 128 bit output block - the 64 bits words
		// are in big endian order but the two 32 bit words within
		// them are in little endian order

		mov	 ebx,res

#ifdef	CHANGE_ORDER

		movq	qword ptr[ebx],mm_ylo
		movq	qword ptr[ebx + 8],mm_yhi
#else
		// reverse endian form to big endian

		movq	qword ptr[ebx],mm_yhi
		movq	qword ptr[ebx + 8],mm_ylo

		mov		eax,dword ptr[ebx]poly128._1
		mov		ecx,dword ptr[ebx]poly128._0
		mov		dword ptr[ebx]poly128._0,eax
		mov		dword ptr[ebx]poly128._1,ecx

		mov		eax,dword ptr[ebx]poly128._3
		mov		ecx,dword ptr[ebx]poly128._2
		mov		dword ptr[ebx]poly128._2,eax
		mov		dword ptr[ebx]poly128._3,ecx
#endif
		emms
	}
}

static mmx64 tab2[256];
static mmx64 tab4[256];

static void MMX_set_tables(void)
{	u32		i;

	for(i = 0; i < 256; ++i)
	{
		tab2[i]._1 = 0; tab2[i]._0 = tgen_2(i);
		tab4[i]._1 = 0; tab4[i]._0 = tgen_4(i);
	}
}

#define mm_yl	mm0
#define mm_yh	mm1
#define mm_t1	mm2
#define mm_t2	mm3
#define mm_t3	mm4

#define	munge_asm				\
	__asm	movq	mm_t2,mm_t1	\
	__asm	psrlq	mm_t2,36	\
	__asm	psllq	mm_t1,19	\
	__asm	pxor	mm_yh,mm_t2	\
	__asm	pxor	mm_yl,mm_t1	\
	__asm	psrlq	mm_t2,7		\
	__asm	psllq	mm_t1,1		\
	__asm	pxor	mm_yh,mm_t2	\
	__asm	pxor	mm_yl,mm_t1	\
	__asm	psrlq	mm_t2,1		\
	__asm	psllq	mm_t1,1		\
	__asm	pxor	mm_yh,mm_t2	\
	__asm	pxor	mm_yl,mm_t1	\
	__asm	psrlq	mm_t2,1		\
	__asm	psllq	mm_t1,7		\
	__asm	pxor	mm_yh,mm_t2	\
	__asm	pxor	mm_yl,mm_t1

#define	xx_0	 0
#define	xx_1	 4
#define	xx_2	 8
#define	xx_3	12

#pragma warning (disable : 4799)

static void MMX_squareNTimes(poly128 *res, poly128 *x, uint n) 
{	u32		xx[6];

	__asm
	{	mov		edx,x
		lea		ebx,xx
		add		ebx,7
		and		ebx,0xfffffff8
		mov		eax,dword ptr[edx]poly128._0
		mov		dword ptr[ebx+xx_0],eax
		mov		ecx,dword ptr[edx]poly128._1
		mov		dword ptr[ebx+xx_1],ecx
		mov		eax,dword ptr[edx]poly128._2
		mov		dword ptr[ebx+xx_2],eax
		mov		ecx,dword ptr[edx]poly128._3
		mov		dword ptr[ebx+xx_3],ecx

		mov		edx,offset tab4
		mov		ecx,n
		cmp		ecx,2
		jb		short lb2
lb1:	mov		eax,dword ptr[ebx+xx_3]
		shr		eax,8
		movq	mm_t1,qword ptr[edx+8*eax]
		mov		eax,dword ptr[ebx+xx_2]
		shr		eax,24
		movq	mm_yh,qword ptr[edx+8*eax]
		mov		eax,dword ptr[ebx+xx_2]
		shr		eax,8
		and		eax,255
		movq	mm_yl,qword ptr[edx+8*eax]
		psllq	mm_t1,32
		psllq	mm_yh,32
		psllq	mm_yl,32
		mov		eax,dword ptr[ebx+xx_3]
		and		eax,255
		pxor	mm_t1,qword ptr[edx+8*eax]
		mov		eax,dword ptr[ebx+xx_2]
		shr		eax,16
		and		eax,255
		pxor	mm_yh,qword ptr[edx+8*eax]
		mov		eax,dword ptr[ebx+xx_2]
		and		eax,255
		pxor	mm_yl,qword ptr[edx+8*eax]
		munge_asm

		movq	mm_t1,mm_yh
		movq	mm_yh,mm_yl			
		mov		eax,dword ptr[ebx+xx_1]
		shr		eax,24
		movq	mm_yl,qword ptr[edx+8*eax]
		psllq	mm_yl,32
		mov		eax,dword ptr[ebx+xx_1]
		shr		eax,16
		and		eax,255
		pxor	mm_yl,qword ptr[edx+8*eax]
		munge_asm

		movq	mm_t1,mm_yh
		movq	mm_yh,mm_yl	
		mov		eax,dword ptr[ebx+xx_1]
		shr		eax,8
		and		eax,255
		movq	mm_yl,qword ptr[edx+8*eax]
		psllq	mm_yl,32
		mov		eax,dword ptr[ebx+xx_1]
		and		eax,255
		pxor	mm_yl,qword ptr[edx+8*eax]
		munge_asm		

		movq	mm_t1,mm_yh
		movq	mm_yh,mm_yl
		mov		eax,dword ptr[ebx+xx_0]
		shr		eax,24
		movq	mm_yl,qword ptr[edx+8*eax]
		psllq	mm_yl, 32;
		mov		eax,dword ptr[ebx+xx_0]
		shr		eax,16
		and		eax,255
		pxor	mm_yl,qword ptr[edx+8*eax]
		munge_asm

		movq	mm_t1,mm_yh
		movq	mm_yh,mm_yl
		mov		eax,dword ptr[ebx+xx_0]
		shr		eax,8
		and		eax,255
		movq	mm_yl,qword ptr[edx+8*eax]
		psllq	mm_yl,32
		mov		eax,dword ptr[ebx+xx_0]
		and		eax,255
		pxor	mm_yl,qword ptr[edx+8*eax]
		munge_asm

		// Reduce modulo t^109+t^9+t^2+t+1
		
		movq	mm_t1,mm_yh 
		psrlq	mm_t1,45
		psllq	mm_yh,19 
		psrlq	mm_yh,19

		pxor	mm_yl,mm_t1
		psllq	mm_t1,1 
		pxor	mm_yl,mm_t1
		psllq	mm_t1,1 
		pxor	mm_yl,mm_t1
		psllq	mm_t1,7 
		pxor	mm_yl,mm_t1

		movq	qword ptr[ebx+xx_2],mm_yh
		movq	qword ptr[ebx+xx_0],mm_yl
		sub		ecx,2
		cmp		ecx,2
		jae		short lb1
		and		ecx,1
		jz		short lb3

lb2:	mov		edx,offset tab2
		mov		eax,dword ptr[ebx+xx_3]
		shr		eax,8
		movq	mm_t2,qword ptr[edx+8*eax]
		mov		eax,dword ptr[ebx+xx_2]
		shr		eax,24
		movq	mm_t1,qword ptr[edx+8*eax]
		mov		eax,dword ptr[ebx+xx_1]
		shr		eax,24
		movq	mm_yh,qword ptr[edx+8*eax]
		mov		eax,dword ptr[ebx+xx_0]
		shr		eax,24
		movq	mm_yl,qword ptr[edx+8*eax]

		psllq	mm_t2,16 
		mov		eax,dword ptr[ebx+xx_3]
		and		eax,255
		por		mm_t2,qword ptr[edx+8*eax]
		psllq	mm_t1,16; 
		mov		eax,dword ptr[ebx+xx_2]
		shr		eax,16
		and		eax,255
		por		mm_t1,qword ptr[edx+8*eax]
		psllq	mm_yh,16
		mov		eax,dword ptr[ebx+xx_1]
		shr		eax,16
		and		eax,255
		por		mm_yh,qword ptr[edx+8*eax]
		psllq	mm_yl,16
		mov		eax,dword ptr[ebx+xx_0]
		shr		eax,16
		and		eax,255
		por		mm_yl,qword ptr[edx+8*eax]
		
		psllq	mm_t1,16
		mov		eax,dword ptr[ebx+xx_2]
		shr		eax,8
		and		eax,255
		por		mm_t1,qword ptr[edx+8*eax]
		psllq	mm_yh,16
		mov		eax,dword ptr[ebx+xx_1]
		shr		eax,8
		and		eax,255
		por		mm_yh,qword ptr[edx+8*eax]
		psllq	mm_yl,16
		mov		eax,dword ptr[ebx+xx_0]
		shr		eax,8
		and		eax,255
		por		mm_yl,qword ptr[edx+8*eax]

		psllq	mm_t1,16
		mov		eax,dword ptr[ebx+xx_2]
		and		eax,255
		por		mm_t1,qword ptr[edx+8*eax]
		psllq	mm_yh,16
		mov		eax,dword ptr[ebx+xx_1]
		and		eax,255
		por		mm_yh,qword ptr[edx+8*eax]
		psllq	mm_yl,16
		mov		eax,dword ptr[ebx+xx_0]
		and		eax,255
		por		mm_yl,qword ptr[edx+8*eax]

		// Reduce modulo t^109+t^9+t^2+t+1

		movq	mm_t3,mm_t1;
		psllq	mm_t2,19  
		pxor	mm_yh,mm_t2
		psrlq	mm_t3,36 
		pxor	mm_yh,mm_t3
		psllq	mm_t1,19  
		pxor	mm_yl,mm_t1

		psllq	mm_t2,1  
		pxor	mm_yh,mm_t2
		psrlq	mm_t3,7 
		pxor	mm_yh,mm_t3
		psllq	mm_t1,1
		pxor	mm_yl,mm_t1

		psllq	mm_t2,1
		pxor	mm_yh,mm_t2
		psrlq	mm_t3,1
		pxor	mm_yh,mm_t3
		psllq	mm_t1,1  
		pxor	mm_yl,mm_t1

		psllq	mm_t2,7
		pxor	mm_yh,mm_t2
		psrlq	mm_t3,1
		pxor	mm_yh,mm_t3
	    psllq	mm_t1,7
		pxor	mm_yl,mm_t1

	    movq	mm_t3,mm_yh
		psrlq	mm_t3,45
		psllq	mm_yh,19 
		psrlq	mm_yh, 19

		pxor	mm_yl,mm_t3
		psllq	mm_t3,1 
		pxor	mm_yl,mm_t3
	    psllq	mm_t3,1 
		pxor	mm_yl,mm_t3
		psllq	mm_t3,7 
		pxor	mm_yl,mm_t3

		movq	qword ptr[ebx+xx_2],mm_yh
		movq	qword ptr[ebx+xx_0],mm_yl
lb3:	emms
		mov		edx,res
		mov		eax,dword ptr[ebx+xx_0]
		mov		ecx,dword ptr[ebx+xx_1]
		mov		dword ptr[edx]poly128._0,eax
		mov		dword ptr[edx]poly128._1,ecx
		mov		eax,dword ptr[ebx+xx_2]
		mov		ecx,dword ptr[ebx+xx_3]
		mov		dword ptr[edx]poly128._2,eax
		mov		dword ptr[edx]poly128._3,ecx
	}
}

// MMX_GF2Product56x56

// Computes 111-bit product of polynomials xh:xl and *py over Z/2Z
// with degree < 56.  Speed-critical auxiliary function used for product()
// Returns with low 64 bits of result in *pl and high 47 bits in *ph.

#define mm_00 mm0
#define mm_01 mm1
#define mm_02 mm2
#define mm_03 mm3
#define mm_04 mm4

#define mm_a  mm0
#define mm_b  mm1
#define mm_c  mm2
#define mm_d  mm3
#define mm_e  mm4
#define mm_f  mm5
#define mm_g  mm6
#define mm_h  mm7

#define	tadr(n)	qword ptr[ebx + 8 * n]

// Thanks to Jim Fougeron for pointing out the need for MMX data alignment on the stack

#pragma warning (disable : 4799)

static void MMX_GF2Product56x56(const mmx64 *px, const mmx64 *py, mmx128 *pr) 
{	mmx64	tab[17];	// one extra to allow alignment adjustment

	// Note: No START_MMX() here.  It is done at start of product()

	__asm
	{	lea		ebx,tab			// align MMX data to 8 byte boundary
		add		ebx,7
		and		ebx,0xfffffff8
		mov		eax,py
		movq	mm_01,qword ptr[eax]
		movq	mm_02,mm_01 
		psllq	mm_02,1
		movq	mm_03, mm_01
		psllq	mm_03,2
		movq	mm_04,mm_01 
		psllq	mm_04,3

		// Gray code walk through table
 
		pxor	mm_00,mm_00  
		movq	tadr(0),mm_00
		movq	mm_00,mm_01 
		movq	tadr(1),mm_01
		pxor	mm_00,mm_02 
		movq	tadr(3),mm_00
		movq	mm_00,mm_02 
		movq	tadr(2),mm_02
		pxor	mm_00,mm_03 
		movq	tadr(6),mm_00
		pxor	mm_00,mm_01 
		movq	tadr(7),mm_00
		pxor	mm_00,mm_02 
		movq	tadr(5),mm_00
		movq	mm_00,mm_03 
		movq	tadr(4),mm_03
		pxor	mm_00,mm_04 
		movq	tadr(12),mm_00
		pxor	mm_00,mm_01 
		movq	tadr(13),mm_00
		pxor	mm_00,mm_02 
		movq	tadr(15),mm_00
		pxor	mm_00,mm_01 
		movq	tadr(14),mm_00
		pxor	mm_00,mm_03 
		movq	tadr(10),mm_00
		pxor	mm_00,mm_01 
		movq	tadr(11),mm_00
		pxor	mm_00,mm_02 
		movq	tadr(9),mm_00
		movq	tadr(8),mm_04

		mov		eax,px
		mov		ecx,dword ptr[eax]mmx64._0
		mov		edx,dword ptr[eax]mmx64._1

		mov		eax,ecx
		shr		eax,1
		and		eax,8*15
		movq	mm_a,qword ptr[ebx+eax]
		mov		eax,ecx
		shr		eax,9
		and		eax,8*15
		movq	mm_b,qword ptr[ebx+eax]
		psllq	mm_a,4
		psllq	mm_b,4
		mov		eax,ecx
		and		eax,15
		pxor	mm_a,qword ptr[ebx+8*eax]
		mov		eax,ecx
		shr		eax,5
		and		eax,8*15
		pxor	mm_b,qword ptr[ebx+eax]

		mov		eax,ecx
		shr		eax,17
		and		eax,8*15
		movq	mm_c,qword ptr[ebx+eax]
		mov		eax,ecx
		shr		eax,28
		movq	mm_d,qword ptr[ebx+8*eax]
		psllq	mm_c,4
		psllq	mm_d,4
		mov		eax,ecx
		shr		eax,13
		and		eax,8*15
		pxor	mm_c,qword ptr[ebx+eax]
		mov		eax,ecx
		shr		eax,21
		and		eax,8*15
		pxor	mm_d,qword ptr[ebx+eax]
		
		mov		eax,edx
		shr		eax,1
		and		eax,8*15
		movq	mm_e,qword ptr[ebx+eax]
		mov		eax,edx
		shr		eax,9
		and		eax,8*15
		movq	mm_f,qword ptr[ebx+eax]
		psllq	mm_e,4
		psllq	mm_f,4
		mov		eax,edx
		and		eax,15
		pxor	mm_e,qword ptr[ebx+8*eax]
		mov		eax,edx
		shr		eax,5
		and		eax,8*15
		pxor	mm_f,qword ptr[ebx+eax]

		mov		eax,edx
		shr		eax,20
		movq	mm_g,qword ptr[ebx+8*eax]
		psllq	mm_g,4
		mov		eax,edx
		shr		eax,13
		and		eax,8*15
		pxor	mm_g,qword ptr[ebx+eax]

		movq	mm_h,mm_b 
		psllq	mm_b,8 
		psrlq	mm_h,8 
		pxor	mm_a,mm_b
		pxor	mm_h,mm_c 
		psllq	mm_c,16 
		psrlq	mm_h,8 
		pxor	mm_a,mm_c
		pxor	mm_h,mm_d 
		psllq	mm_d,24 
		psrlq	mm_h,8 
		pxor	mm_a,mm_d
		pxor	mm_h,mm_e 
		psllq	mm_e,32 
		psrlq	mm_h,8 
		pxor	mm_a,mm_e
		pxor	mm_h,mm_f 
		psllq	mm_f,40 
		psrlq	mm_h,8 
		pxor	mm_a,mm_f
		pxor	mm_h,mm_g 
		psllq	mm_g,48 
		psrlq	mm_h,16 
		pxor	mm_a,mm_g
		mov		eax,pr
		movq	qword ptr[eax]mmx128._0,mm_a
		movq	qword ptr[eax]mmx128._1,mm_h
	}
	// Note: No END_MMX() here.  It is done at end of product()
}

#define mm_u0 mm0
#define mm_u1 mm1
#define mm_u2 mm2
#define mm_u3 mm3
#define mm_u4 mm4
#define mm_mh mm5
#define mm_ml mm6

static void MMX_product(poly128 *x, const poly128 *y) 
{	mmx64		xx,yy;
	mmx128		r_2, r_1, r_0;
	const u32	mask = 0x00ffffff;

	xx._1 = x->_1 & mask; xx._0 = x->_0;
	yy._1 = y->_1 & mask; yy._0 = y->_0;
    MMX_GF2Product56x56(&xx, &yy, &r_0);

    xx._1 = (x->_3 << 8 | x->_2 >> 24) & mask; xx._0 = x->_2 << 8 | x->_1 >> 24;
    yy._1 = (y->_3 << 8 | y->_2 >> 24) & mask; yy._0 = y->_2 << 8 | y->_1 >> 24;
    MMX_GF2Product56x56(&xx, &yy, &r_2);

	xx._1 ^= x->_1 & mask; xx._0 ^= x->_0;
    yy._1 ^= y->_1 & mask; yy._0 ^= y->_0; 
    MMX_GF2Product56x56(&xx, &yy, &r_1);
    
	__asm
	{	lea		ebx,r_0
		lea		ecx,r_1
		lea		edx,r_2

		movq	mm_u0, qword ptr[ebx]mmx128._0
		movq	mm_u2, qword ptr[edx]mmx128._0
		movq	mm_u3, qword ptr[edx]mmx128._1
		movq	mm_u1, qword ptr[ebx]mmx128._1
		movq	mm_ml, mm_u2
		movq	mm_mh, mm_u3
		psrlq	mm_u2,16
		psrlq	mm_u3,16
		psllq	mm_ml,48
		psllq	mm_mh,48
		por		mm_u1,mm_ml
		por		mm_u2,mm_mh
		movq	mm_ml,qword ptr[ecx]mmx128._0
		movq	mm_mh,qword ptr[ecx]mmx128._1
		pxor	mm_ml,qword ptr[ebx]mmx128._0
		pxor	mm_mh,qword ptr[ebx]mmx128._1
		pxor	mm_ml,qword ptr[edx]mmx128._0
		pxor	mm_mh,qword ptr[edx]mmx128._1

		movq	mm_u4,mm_ml 
		psllq	mm_u4,56 
		pxor	mm_u0,mm_u4
		psrlq	mm_ml,8 
		pxor	mm_u1,mm_ml
		movq	mm_u4,mm_mh 
		psllq	mm_u4,56 
		pxor	mm_u1,mm_u4
		psrlq	mm_mh,8 
		pxor	mm_u2, mm_mh

		// Reduce modulo t^109+t^9+t^2+t+1
	
		movq	mm_u4,mm_u2
		psllq	mm_u3,19  
		pxor	mm_u1,mm_u3
		psrlq	mm_u4,36 
		pxor	mm_u1,mm_u4
		psllq	mm_u2,19  
		pxor	mm_u0,mm_u2

		psllq	mm_u3,1  
		pxor	mm_u1,mm_u3
		psrlq	mm_u4,7 
		pxor	mm_u1,mm_u4
		psllq	mm_u2,1  
		pxor	mm_u0,mm_u2

		psllq	mm_u3,1  
		pxor	mm_u1,mm_u3
		psrlq	mm_u4,1 
		pxor	mm_u1,mm_u4
		psllq	mm_u2,1  
		pxor	mm_u0,mm_u2

		psllq	mm_u3,7  
		pxor	mm_u1,mm_u3
		psrlq	mm_u4,1
		pxor	mm_u1,mm_u4
		psllq	mm_u2,7  
		pxor	mm_u0,mm_u2

		movq	mm_u4,mm_u1 
		psrlq	mm_u4,45
		psllq	mm_u1,19 
		psrlq	mm_u1,19

		pxor	mm_u0,mm_u4
		psllq	mm_u4,1 
		pxor	mm_u0,mm_u4
		psllq	mm_u4,1 
		pxor	mm_u0,mm_u4
		psllq	mm_u4,7 
		pxor	mm_u0,mm_u4

		// set result in 128 bit output block - the 64 bits words
		// are in big endian order but the two 32 bit words within
		// them are in little endian order

		mov		ebx,x

#ifdef CHANGE_ORDER

		movq	qword ptr[ebx],mm_u0			// least sig 64 bits
		movq	qword ptr[ebx + 8],mm_u1		// most sig 64 bits
#else	
		// reverse endian form to big endian

		movq	qword ptr[ebx],mm_u1			// most sig 64 bits
		movq	qword ptr[ebx + 8],mm_u0		// least sig 64 bits

		mov		eax,dword ptr[ebx]poly128._0
		mov		ecx,dword ptr[ebx]poly128._1
		mov		dword ptr[ebx]poly128._0,ecx
		mov		dword ptr[ebx]poly128._1,eax

		mov		eax,dword ptr[ebx]poly128._2
		mov		ecx,dword ptr[ebx]poly128._3
		mov		dword ptr[ebx]poly128._2,ecx
		mov		dword ptr[ebx]poly128._3,eax
#endif
		emms		
	}
}

// end of assembler coded MMX based functions

// p and q point to mmx ordered data so that this routine
// can use the same table as the MMX code above

#define FOR_BLOCK									\
{	q = &p[t & 3]; yt ^= q->_1._1;					\
	yh ^= q->_1._0; ym ^= q->_0._1; yl ^= q->_0._0;	\
    t >>= 2; p += 4;								\
    q = &p[t & 3]; yt ^= q->_1._1;					\
	yh ^= q->_1._0; ym ^= q->_0._1; yl ^= q->_0._0;	\
    t >>= 2; p += 4;								\
}

static void nonMMX_change_base(poly128 *res, const poly128 *x) 
{	uint			i;
	u32				t,yt,yh,ym,yl;
	const mmx128	*p, *q;

	yt = yh = ym = yl = 0;
  
	p = (const mmx128*)&matrix[0];

	t = x->_0; for(i = 8; i; --i) FOR_BLOCK
	t = x->_1; for(i = 8; i; --i) FOR_BLOCK
	t = x->_2; for(i = 8; i; --i) FOR_BLOCK
	t = x->_3; for(i = 3; i; --i) FOR_BLOCK

	q = &p[t & 1]; 
	yt ^= q->_1._1; yh ^= q->_1._0; 
	ym ^= q->_0._1; yl ^= q->_0._0;

    res->_3 = yt; res->_2 = yh; res->_1 = ym; res->_0 = yl;
}

static u32 tab[256];

static void nonMMX_set_tables(void)
{	u32		i;
	
	for(i = 0; i < 256; ++i)
	
		tab[i] = tgen_2(i);
}

static void nonMMX_squareNTimes(poly128 *res, poly128 *x, uint n) 
{	u32		xx[4];

	xx[3] = x->_3; xx[2] = x->_2; xx[1] = x->_1; xx[0] = x->_0;

	for( ; n; --n) 
	{
		u32		t0,t1,t2,t3,t4,t5,t6, w;

		t6 = tab[xx[3] & 255]         | tab[(xx[3] >> 8) & 255] << 16;
		t5 = tab[(xx[2] >> 16) & 255] | tab[xx[2] >> 24] << 16;
		t4 = tab[xx[2] & 255]         | tab[(xx[2] >> 8) & 255] << 16;
		t3 = tab[(xx[1] >> 16) & 255] | tab[xx[1] >> 24] << 16;
		t2 = tab[xx[1] & 255]         | tab[(xx[1] >> 8) & 255] << 16;
		t1 = tab[(xx[0] >> 16) & 255] | tab[xx[0] >> 24] << 16;
		t0 = tab[xx[0] & 255]         | tab[(xx[0] >> 8) & 255] << 16;

		// Reduce modulo t^109+t^9+t^2+t+1

		t3 ^= t6 >> 13; t3 ^= t6 >> 12; t3 ^= t6 >> 11; t3 ^= t6 >>  4;
		t2 ^= t6 << 19; t2 ^= t6 << 20; t2 ^= t6 << 21; t2 ^= t6 << 28;
		t2 ^= t5 >> 13; t2 ^= t5 >> 12; t2 ^= t5 >> 11; t2 ^= t5 >>  4;
		t1 ^= t5 << 19; t1 ^= t5 << 20; t1 ^= t5 << 21; t1 ^= t5 << 28;
		t1 ^= t4 >> 13; t1 ^= t4 >> 12; t1 ^= t4 >> 11; t1 ^= t4 >>  4;
		t0 ^= t4 << 19; t0 ^= t4 << 20; t0 ^= t4 << 21; t0 ^= t4 << 28;

		w = t3 >> 13; t3 ^= w << 13; t0 ^= w; t0 ^= w<<1; t0 ^= w<<2; t0 ^= w << 9;
		xx[3] = t3; xx[2] = t2; xx[1] = t1; xx[0] = t0;
	}

	res->_3 = xx[3]; res->_2 = xx[2]; res->_1 = xx[1]; res->_0 = xx[0];
}

#if PROD == 1

// GF2Product27x27

// Computes 53-bit product of polynomials over Z/2Z with degree < 27
// Returns low 32 bits of result, and leaves top 21 bits in *ph

static u32 GF2Product27x27(u32 a, u32 b, u32 *ph) 
{	u32		a2, a4, h, l;

	a2 = a << 1; a4 = a << 2;

	{	u32 tab[8] = { 0, a, a2, a ^ a2, a4, a ^ a4, a2 ^ a4, a ^ a2 ^ a4 };

#if 0
		l = tab[b>>24];
		l = l <<  3 ^ tab[(b >> 21) & 7];
		h = l >> 29; l = (l << 3) ^ tab[(b >> 18) & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[(b >> 15) & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[(b >> 12) & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[(b >>  9) & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[(b >>  6) & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[(b >>  3) & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[b & 7];
#else
		u32 s, t;

		s = tab[b     & 7];
		t = tab[(b >>  3) & 7]; l = s;
		s = tab[(b >>  6) & 7]; l ^= t <<  3;
		t = tab[(b >>  9) & 7]; l ^= s <<  6; h  = s >> 26;
		s = tab[(b >> 12) & 7]; l ^= t <<  9; h ^= t >> 23;
		t = tab[(b >> 15) & 7]; l ^= s << 12; h ^= s >> 20;
		s = tab[(b >> 18) & 7]; l ^= t << 15; h ^= t >> 17;
		t = tab[(b >> 21) & 7]; l ^= s << 18; h ^= s >> 14;
		s = tab[(b >> 24)    ]; l ^= t << 21; h ^= t >> 11;
						    	l ^= s << 24; h ^= s >>  8;
#endif
	    *ph = h; return l;
	}
}

#elif PROD == 3

// GF2Product28x28

// Computes 55-bit product of polynomials over Z/2Z with degree < 28
// Returns low 32 bits of result, and leaves top 23 bits in *ph

static u32 GF2Product28x28(u32 a, u32 b, u32 *ph) 
{	u32		a2, a4, h, l;

	a2 = a << 1; a4 = a << 2;

	{	u32 tab[8] = { 0, a, a2, a ^ a2, a4, a ^ a4, a2 ^ a4, a ^ a2 ^ a4 };

#if 0
		l = tab[b>>27];
		l = l <<  3 ^ tab[b>>24 & 7];
		h = l >> 29; l = l<<3 ^ tab[b>>21 & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[b >> 18 & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[b >> 15 & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[b >> 12 & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[b >>  9 & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[b >>  6 & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[b >>  3 & 7];
		h = h <<  3 | l >> 29; l = l << 3 ^ tab[b & 7];
#else
		u32 s, t;

		s = tab[b     & 7];
		t = tab[b >>  3 & 7]; l = s;
		s = tab[b >>  6 & 7]; l ^= t <<  3; h  = t >> 29;
		t = tab[b >>  9 & 7]; l ^= s <<  6; h ^= s >> 26;
		s = tab[b >> 12 & 7]; l ^= t <<  9; h ^= t >> 23;
		t = tab[b >> 15 & 7]; l ^= s << 12; h ^= s >> 20;
		s = tab[b >> 18 & 7]; l ^= t << 15; h ^= t >> 17;
		t = tab[b >> 21 & 7]; l ^= s << 18; h ^= s >> 14;
		s = tab[b >> 24 & 7]; l ^= t << 21; h ^= t >> 11;
		t = tab[b >> 27    ]; l ^= s << 24; h ^= s >>  8;
							  l ^= t << 27; h ^= t >>  5;
#endif
	    *ph = h; return l;
	}
}

#elif PROD == 5

// GF2Product32x32

// Computes 63-bit product of polynomials over Z/2Z with degree < 32.
// Returns low 32 bits of result, and leaves top 31 bits in *ph

static u32 GF2Product32x32(u32 a, u32 b, u32 *ph) 
{	u32 a1,a2,a4, h, l;

	a4 = a << 2; a2 = a4 >> 1; a1 = a4 >> 2;

	{	u32 s, t, tab[8] = { 0, a1, a2, a1 ^ a2, a4, a1 ^ a4, a2 ^ a4, a1 ^ a2 ^ a4 };

#if 0
		l = h = 0;
		if((a >> 31) & 1) { l =  b << 31; h =  b >> 1; }
		if((a >> 30) & 1) { l ^= b << 30; h ^= b >> 2; }

		s = tab[b       & 7];
		t = tab[b >>  3 & 7]; l ^= s;
		s = tab[b >>  6 & 7]; l ^= t <<  3; h ^= t >> 29;
		t = tab[b >>  9 & 7]; l ^= s <<  6; h ^= s >> 26;
		s = tab[b >> 12 & 7]; l ^= t <<  9; h ^= t >> 23;
		t = tab[b >> 15 & 7]; l ^= s << 12; h ^= s >> 20;
		s = tab[b >> 18 & 7]; l ^= t << 15; h ^= t >> 17;
		t = tab[b >> 21 & 7]; l ^= s << 18; h ^= s >> 14;
		s = tab[b >> 24 & 7]; l ^= t << 21; h ^= t >> 11;
		t = tab[b >> 27 & 7]; l ^= s << 24; h ^= s >>  8;
		s = tab[b >> 30    ]; l ^= t << 27; h ^= t >>  5;
							  l ^= s << 30; h ^= s >>  2;
#else
		s = tab[b       & 7];
		t = tab[b >>  3 & 7]; l = s;
		s = tab[b >>  6 & 7]; l ^= t <<  3; h  = t >> 29;
		t = tab[b >>  9 & 7]; l ^= s <<  6; h ^= s >> 26;
		s = tab[b >> 12 & 7]; l ^= t <<  9; h ^= t >> 23;
		t = tab[b >> 15 & 7]; l ^= s << 12; h ^= s >> 20;
		s = tab[b >> 18 & 7]; l ^= t << 15; h ^= t >> 17;
		t = tab[b >> 21 & 7]; l ^= s << 18; h ^= s >> 14;
		s = tab[b >> 24 & 7]; l ^= t << 21; h ^= t >> 11;
		t = tab[b >> 27 & 7]; l ^= s << 24; h ^= s >>  8;
		s = tab[b >> 30    ]; l ^= t << 27; h ^= t >>  5;
							  l ^= s << 30; h ^= s >>  2;

		if((a >> 31) & 1) { l ^= b << 31; h ^= b >> 1; }
		if((a >> 30) & 1) { l ^= b << 30; h ^= b >> 2; }
#endif
	}

	*ph = h; return l;
}

#endif

#define STASH(t)	tab[2 * t] = w0; tab[2 * t + 1] = w1;
#define GRAB(x,n)	t = (x >> (4 * n)) & 15; w0 = tab[2 * t]; w1 = tab[2 * t + 1];

#if PROD == 1

// GF2Product54x54

// Computes 107-bit product of polynomials xh:xl and yh:yl over Z/2Z
// with degree < 54.  Speed-critical auxiliary function used for product()
// Returns low 32 bits of result, leaves middle 64 bits in *pm and *ph
// and high 11 bits in *pt

static u32 GF2Product54x54(u32 xh, u32 xl, u32 yh, u32 yl, u32 *pt, u32 *ph, u32 *pm) 
{	u32			l0, l1, m0, m1, h0, h1;
	const u32	mask = (1ul << 27 ) - 1;

	xh = (xh << 5) | (xl >> 27); xl &= mask;
	yh = (yh << 5) | (yl >> 27); yl &= mask;

	h0 = GF2Product27x27(xh, yh, &h1);
	l0 = GF2Product27x27(xl, yl, &l1);
	m0 = GF2Product27x27(xl ^ xh, yl ^ yh, &m1);

	m0 ^= l0; m1 ^= l1; m0 ^= h0; m1 ^= h1;
	l1 ^= h0 << 22; h0  = (h0 >> 10) | (h1 << 22); h1 >>= 10;
	l0 ^= m0 << 27; l1 ^= (m0 >>  5) | (m1 << 27); h0 ^= m1 >> 5;

	*pt = h1; *ph = h0; *pm = l1; return l0;
}

#elif PROD == 2

static u32 GF2Product54x54(u32 xh, u32 xl, u32 yh, u32 yl, u32 *pt, u32 *ph, u32 *pm) 
{	u32 tab[32],w0,w1,yh2,yh4,yh8,a0,a1,a2,a3,t;

	yh2 = (yh << 1) ^ (yl >> 31);
	yh4 = (yh << 2) ^ (yl >> 30);
	yh8 = (yh << 3) ^ (yl >> 29);

	// Gray code walk through table

	w0  = 0;            w1  = 0;         STASH( 0);
	w0  = yl;           w1  = yh;        STASH( 1);
	w0  = yl ^ yl << 1; w1  = yh ^ yh2;  STASH( 3);
	w0  = yl << 1;      w1  = yh2;       STASH( 2);
	w0 ^= yl << 2;      w1  = yh2 ^ yh4; STASH( 6);
	w0 ^= yl;           w1 ^= yh;        STASH( 7);
	w0  = yl ^ yl << 2; w1  = yh ^ yh4;  STASH( 5);
	w0  = yl << 2;      w1  = yh4;       STASH( 4);
	w0 ^= yl << 3;      w1  = yh4 ^ yh8; STASH(12);
	w0 ^= yl;           w1 ^= yh;        STASH(13);
	w0 ^= yl << 1;      w1 ^= yh2;       STASH(15);
	w0 ^= yl;           w1 ^= yh;        STASH(14);
	w0 ^= yl << 2;      w1  = yh2 ^ yh8; STASH(10);
	w0 ^= yl;           w1 ^= yh;        STASH(11);
	w0  = yl ^ yl << 3; w1  = yh ^ yh8;  STASH( 9);
	w0  = yl << 3;      w1  = yh8;       STASH( 8);

	GRAB(xl, 0); a0 =  w0;       a1 =  w1;
	GRAB(xl, 1); a0 ^= w0 <<  4; a1 ^= w0 >> 28; a1 ^= w1 <<  4;
	GRAB(xl, 2); a0 ^= w0 <<  8; a1 ^= w0 >> 24; a1 ^= w1 <<  8; a2  = w1 >> 24;
	GRAB(xl, 3); a0 ^= w0 << 12; a1 ^= w0 >> 20; a1 ^= w1 << 12; a2 ^= w1 >> 20;
	GRAB(xl, 4); a0 ^= w0 << 16; a1 ^= w0 >> 16; a1 ^= w1 << 16; a2 ^= w1 >> 16;
	GRAB(xl, 5); a0 ^= w0 << 20; a1 ^= w0 >> 12; a1 ^= w1 << 20; a2 ^= w1 >> 12;
	GRAB(xl, 6); a0 ^= w0 << 24; a1 ^= w0 >>  8; a1 ^= w1 << 24; a2 ^= w1 >>  8;
	GRAB(xl, 7); a0 ^= w0 << 28; a1 ^= w0 >>  4; a1 ^= w1 << 28; a2 ^= w1 >>  4;

	GRAB(xh, 0); a1 ^= w0;       a2 ^= w1;
	GRAB(xh, 1); a1 ^= w0 <<  4; a2 ^= w0 >> 28; a2 ^= w1 <<  4;
	GRAB(xh, 2); a1 ^= w0 <<  8; a2 ^= w0 >> 24; a2 ^= w1 <<  8; a3  = w1 >> 24;
	GRAB(xh, 3); a1 ^= w0 << 12; a2 ^= w0 >> 20; a2 ^= w1 << 12; a3 ^= w1 >> 20;
	GRAB(xh, 4); a1 ^= w0 << 16; a2 ^= w0 >> 16; a2 ^= w1 << 16; a3 ^= w1 >> 16;
	GRAB(xh, 5); a1 ^= w0 << 20; a2 ^= w0 >> 12; a2 ^= w1 << 20; a3 ^= w1 >> 12;

	*pt = a3; *ph = a2; *pm = a1; return a0;
}

#elif PROD == 3

// GF2Product56x56

// Computes 111-bit product of polynomials xh:xl and yh:yl over Z/2Z
// with degree < 56.  Speed-critical auxiliary function used for product()
// Returns low 32 bits of result, leaves middle 64 bits in *pm and *ph
// and high 15 bits in *pt.

static u32 GF2Product56x56(u32 xh, u32 xl, u32 yh, u32 yl, u32 *pt, u32 *ph, u32 *pm) 
{	u32			l0,l1, m0,m1, h0,h1;
	const u32	mask = (1ul << 28) - 1;

	xh = (xh << 4) | (xl >> 28); xl &= mask;
	yh = (yh << 4) | (yl >> 28); yl &= mask;

	h0 = GF2Product28x28(xh, yh, &h1);
	l0 = GF2Product28x28(xl, yl, &l1);
	m0 = GF2Product28x28(xl ^ xh, yl ^ yh, &m1);

	m0 ^= l0; m1 ^= l1; m0 ^= h0; m1 ^= h1;
	l1 ^= h0 << 24; h0  = (h0 >> 8) | (h1 << 24); h1 >>= 8;
	l0 ^= m0 << 28; l1 ^= (m0 >> 4) | (m1 << 28); h0 ^= m1 >> 4;

	*pt = h1; *ph = h0; *pm = l1; return l0;
}

#elif PROD == 4

static u32 GF2Product56x56(u32 xh, u32 xl, u32 yh, u32 yl, u32 *pt, u32 *ph, u32 *pm) 
{	u32	tab[32],w0,w1,yh2,yh4,yh8, a0,a1,a2,a3,t;

    yh2 = (yh << 1) ^ (yl >> 31);
	yh4 = (yh << 2) ^ (yl >> 30);
	yh8 = (yh << 3) ^ (yl >> 29);

	// Gray code walk through table

	w0  = 0;            w1  = 0;         STASH( 0);
	w0  = yl;           w1  = yh;        STASH( 1);
	w0  = yl ^ yl << 1; w1  = yh ^ yh2;  STASH( 3);
	w0  = yl << 1;      w1  = yh2;       STASH( 2);
	w0 ^= yl << 2;      w1  = yh2 ^ yh4; STASH( 6);
	w0 ^= yl;           w1 ^= yh;        STASH( 7);
	w0  = yl ^ yl << 2; w1  = yh ^ yh4;  STASH( 5);
	w0  = yl << 2;      w1  = yh4;       STASH( 4);
	w0 ^= yl << 3;      w1  = yh4 ^ yh8; STASH(12);
	w0 ^= yl;           w1 ^= yh;        STASH(13);
	w0 ^= yl << 1;      w1 ^= yh2;       STASH(15);
	w0 ^= yl;           w1 ^= yh;        STASH(14);
	w0 ^= yl << 2;      w1  = yh2 ^ yh8; STASH(10);
	w0 ^= yl;           w1 ^= yh;        STASH(11);
	w0  = yl ^ yl << 3; w1  = yh ^ yh8;  STASH( 9);
	w0  = yl << 3;      w1  = yh8;       STASH( 8);

	GRAB(xl, 0); a0  = w0;       a1  = w1;
	GRAB(xl, 1); a0 ^= w0 <<  4; a1 ^= w0 >> 28; a1 ^= w1 <<  4;
	GRAB(xl, 2); a0 ^= w0 <<  8; a1 ^= w0 >> 24; a1 ^= w1 <<  8; a2  = w1 >> 24;
	GRAB(xl, 3); a0 ^= w0 << 12; a1 ^= w0 >> 20; a1 ^= w1 << 12; a2 ^= w1 >> 20;
	GRAB(xl, 4); a0 ^= w0 << 16; a1 ^= w0 >> 16; a1 ^= w1 << 16; a2 ^= w1 >> 16;
	GRAB(xl, 5); a0 ^= w0 << 20; a1 ^= w0 >> 12; a1 ^= w1 << 20; a2 ^= w1 >> 12;
	GRAB(xl, 6); a0 ^= w0 << 24; a1 ^= w0 >>  8; a1 ^= w1 << 24; a2 ^= w1 >>  8;
	GRAB(xl, 7); a0 ^= w0 << 28; a1 ^= w0 >>  4; a1 ^= w1 << 28; a2 ^= w1 >>  4;

	GRAB(xh, 0); a1 ^= w0;       a2 ^= w1;
	GRAB(xh, 1); a1 ^= w0 <<  4; a2 ^= w0 >> 28; a2 ^= w1 <<  4;
	GRAB(xh, 2); a1 ^= w0 <<  8; a2 ^= w0 >> 24; a2 ^= w1 <<  8; a3  = w1 >> 24;
	GRAB(xh, 3); a1 ^= w0 << 12; a2 ^= w0 >> 20; a2 ^= w1 << 12; a3 ^= w1 >> 20;
	GRAB(xh, 4); a1 ^= w0 << 16; a2 ^= w0 >> 16; a2 ^= w1 << 16; a3 ^= w1 >> 16;
	GRAB(xh, 5); a1 ^= w0 << 20; a2 ^= w0 >> 12; a2 ^= w1 << 20; a3 ^= w1 >> 12;

	*pt = a3; *ph = a2; *pm = a1; return a0;
}

#else // PROD == 5

static u32 GF2Product56x56(u32 xh, u32 xl, u32 yh, u32 yl, u32 *pt, u32 *ph, u32 *pm) 
{	u32		l0,l1,m0,m1,h0,h1,t;

	m0 = GF2Product32x32(xl ^ xh, yl ^ yh, &m1);
	l0 = GF2Product32x32(xl, yl, &l1);
	h0 = GF2Product32x32(xh, yh, &h1);

	t = l1 ^ h0; l1 = m0 ^ l0 ^ t; h0 = m1 ^ h1 ^ t;

	*pt = h1; *ph = h0; *pm = l1; return l0;
}

#endif

// Product of x and y in the field GF(2^109) i.e., as polys in (Z/2Z)[t]
// reduced modulo t^109+t^9+t^2+t+1.  Degree x, y < 109

static void nonMMX_product(poly128 *x, const poly128 *y) 
{	u32		lt,lh,lm,ll,mt,mh,mm,ml,ht,hh,hm,hl,t0,t1,t2,t3,t4,t5,t6, w;
	u32		mask,xt,xh,yt,yh;

#if PROD == 1 || PROD == 2

	mask = (1ul << 22) - 1;

    ll = GF2Product54x54(x->_1 & mask, x->_0, y->_1 & mask, y->_0, &lt, &lh, &lm);

    xt = (x->_3 << 10) | (x->_2 >> 22); xh = (x->_2 << 10) | (x->_1 >> 22);
    yt = (y->_3 << 10) | (y->_2 >> 22); yh = (y->_2 << 10) | (y->_1 >> 22);

    ml = GF2Product54x54((xt ^ x->_1) & mask, xh ^ x->_0, (yt ^ y->_1) & mask, yh ^ y->_0, &mt, &mh, &mm);
    hl = GF2Product54x54(xt & mask, xh, yt & mask, yh, &ht, &hh, &hm);

	ml ^= hl; mm ^= hm; mh ^= lh; mt ^= lt;
	ml ^= ll; mm ^= lm; mh ^= hh; mt ^= ht;

	t0 = ll; t1 = lm; t2 = lh; t3 = lt | hl<<12;
	t4 = (hl >> 20) | (hm << 12); t5 = (hm >> 20) | (hh << 12); t6 = (hh >> 20) | (ht << 12);
	t1 ^= ml << 22; t2 ^= (ml >> 10) | (mm << 22); t3 ^= (mm >> 10) | (mh << 22);
	t4 ^= (mh >> 10) | (mt << 22); t5 ^= mt >> 10;

	// Handle 109th bit (TO DO: the "& 1" is not necessary)
	
	if((x->_3 >> 12) & 1) 
	{
		t3 ^= y->_0 << 12; t4 ^= (y->_0 >> 20) | (y->_1 << 12);
		t5 ^= (y->_1 >> 20) | (y->_2 << 12); t6 ^= (y->_2 >> 20) | (y->_3 << 12);
	}

	if((y->_3 >> 12) & 1) 
	{
		t3 ^= x->_0 << 12; t4 ^= (x->_0 >> 20) | (x->_1 << 12);
		t5 ^= (x->_1 >> 20) | (x->_2 << 12); t6 ^= (x->_2 >> 20) | (x->_3 << 12);
	} 
	
	if(((x->_3 & y->_3) >> 12) & 1) 
		
		t6 ^= 1ul << 24;

#else // PROD == 3 || PROD == 4 || PROD == 5 

	mask = (1ul << 24) - 1;

    xt = (x->_3 << 8) | (x->_2 >> 24); xh = (x->_2 << 8) | (x->_1 >> 24);
    yt = (y->_3 << 8) | (y->_2 >> 24); yh = (y->_2 << 8) | (y->_1 >> 24);

    ml = GF2Product56x56((xt ^ x->_1) & mask, xh ^ x->_0, (yt ^ y->_1) & mask, 
														yh ^ y->_0, &mt, &mh, &mm);
    hl = GF2Product56x56(xt & mask, xh, yt & mask, yh, &ht, &hh, &hm);
    ll = GF2Product56x56(x->_1 & mask, x->_0, y->_1 & mask, y->_0, &lt, &lh, &lm);

	ml ^= ll; mm ^= lm; mh ^= hh; mt ^= ht;
	ml ^= hl; mm ^= hm; mh ^= lh; mt ^= lt;

	t0 = ll; t1 = lm; t2 = lh; t3 = lt | (hl << 16);
	t4 = (hl >> 16) | (hm << 16); t5 = (hm >> 16) | (hh << 16); t6 = (hh >> 16) | (ht << 16);
	t1 ^= ml << 24; t2 ^= (ml >> 8) | (mm << 24); t3 ^= (mm >> 8) | (mh << 24);
	t4 ^= (mh >> 8) | (mt << 24); t5 ^= mt >> 8;

#endif // PROD == 1 || PROD == 2 ... #else

  // Reduce modulo t^109+t^9+t^2+t+1

	t3 ^= t6 >> 13; t3 ^= t6 >> 12; t3 ^= t6 >> 11; t3 ^= t6 >>  4;
	t2 ^= t6 << 19; t2 ^= t6 << 20; t2 ^= t6 << 21; t2 ^= t6 << 28;
	t2 ^= t5 >> 13; t2 ^= t5 >> 12; t2 ^= t5 >> 11; t2 ^= t5 >>  4;
	t1 ^= t5 << 19; t1 ^= t5 << 20; t1 ^= t5 << 21; t1 ^= t5 << 28;
	t1 ^= t4 >> 13; t1 ^= t4 >> 12; t1 ^= t4 >> 11; t1 ^= t4 >>  4;
	t0 ^= t4 << 19; t0 ^= t4 << 20; t0 ^= t4 << 21; t0 ^= t4 << 28;

	w = t3 >> 13; t3 ^= w << 13; t0 ^= w; t0 ^= w << 1; t0 ^= w << 2; t0 ^= w << 9;

	x->_3 = t3; x->_2 = t2; x->_1 = t1; x->_0 = t0;
}

// End of C coded versions of MMX based functions

// Detect if MMX instructions are available and set up the 
// appropriate funtion calls depending on the result. Also
// initialise atables and the random number generator.

typedef	void (*pf_1)(poly128*, const poly128*); 
typedef	void (*pf_2)(poly128*, poly128*, uint); 

pf_1	change_base;
pf_2	squareNTimes;
pf_1	product;

static void init_program(void)
{
	MMX_set_tables(); nonMMX_set_tables();

	if(has_mmx())
	{
		change_base = MMX_change_base;
		squareNTimes = MMX_squareNTimes;
		product = MMX_product;
	}
	else
	{
		change_base = nonMMX_change_base;
		squareNTimes = nonMMX_squareNTimes;
		product = nonMMX_product;
	}
}

// SHR(a0, a1, a2, a3, sh, th) shifts a3:a2:a1:a0 right sh bits (th = 32 - sh)
// SHR1(a0, a1, a2, a3) shifts a3:a2:a1:a0 right by one bit
// SHR1_BIS(a0, a1, a2, a3) is similar but discards a3 at the end

#define SHR_BIS(a0, a1, a2, a3, sh, th) \
{										\
    a0 = a0 >> sh | a1 << th;		    \
    a1 = a1 >> sh | a2 << th;			\
    a2 = a2 >> sh | a3 << th;			\
}

#define SHR(a0, a1, a2, a3, sh, th)		\
{										\
	SHR_BIS(a0, a1, a2, a3, sh, th)		\
    a3 >>= sh;							\
}

#define SHR1(a0, a1, a2, a3)	 SHR(a0, a1, a2, a3, 1, 31)
#define SHR1_BIS(a0, a1, a2, a3) SHR_BIS(a0, a1, a2, a3, 1, 31)

// Invert y in the field GF(2^109) i.e., as a poly in (Z/2Z)[t]
// reduced modulo t^109+t^9+t^2+t+1.  Degree y < 109, y != 0
// Note: 1/(1+t+t^2+t^9+O(t^10)) = 1+t+t^3+t^4+t^6+t^7+O(t^10)

static const u8 shiftTab[64] =
{ 6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
  5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0
};

static const u8 xorTab[64] =
{  0, 27, 54, 45, 44, 55, 26,  1, 24,  3, 46, 53, 52, 47,  2, 25, 48, 43,
   6, 29, 28,  7, 42, 49, 40, 51, 30,  5,  4, 31, 50, 41, 32, 59, 22, 13,
  12, 23, 58, 33, 56, 35, 14, 21, 20, 15, 34, 57, 16, 11, 38, 61, 60, 39,
  10, 17,  8, 19, 62, 37, 36, 63, 18,  9
};

static const u32 xorTab2[64] =
{     0, 13889, 27778, 23235, 22724, 28293, 13382,   519, 12360,  1545,
  23754, 27275, 26764, 24269,  1038, 12879, 24720, 22225,  3090, 14931,
  14420,  3605, 21718, 25239, 20696, 26265, 15450,  2587,  2076, 15965,
  25758, 21215, 16608, 30369, 11362,  6691,  6180, 11877, 29862, 17127,
  28840, 18153,  7210, 10859, 10348,  7725, 17646, 29359,  8304,  5681,
  19698, 31411, 30900, 20213,  5174,  8823,  4152,  9849, 31930, 19195,
  18684, 32445,  9342,  4671
};

static void inverse(poly128 *res, const poly128 *y) 
{	uint		sh, th;
	u32			al,am,ah,at, bl,bm,bh,bt, t, ul,um,uh,ut, vl,vm,vh,vt;
	const u32	mt = 1ul << 13, ml = 519; // Note: mm and mh are 0

	// Maintain: u->y = a and v->y = b modulo t^109+t^9+t^2+t+1

	al = y->_0; am = y->_1; ah = y->_2; at = y->_3;
	ul = 1; um = uh = ut = 0;

	while(!(al & 1)) 
	{
		t = ul & 1; ut ^= t << 13; if(t) ul ^= ml;
		
		SHR1(al, am, ah, at); SHR1(ul, um, uh, ut);
	}

	bl = ml; bm = bh = 0; bt = mt; vl = vm = vh = vt = 0;

	do 
	{	do 
		{	do 
			{	do 
				{	do 
					{
						bl ^= al; bm ^= am; bh ^= ah; bt ^= at;
						vl ^= ul; vm ^= um; vh ^= uh; vt ^= ut;

						sh = shiftTab[bl & 63]; th = 32 - sh;
						t = vl & 63; vt ^= (u32)xorTab[t] << 13; 
						vl ^= xorTab2[t];
						SHR(bl, bm, bh, bt, sh, th);
						SHR(vl, vm, vh, vt, sh, th);
            
						while(!(bl & 1)) 
						{
							t = vl & 1; vt ^= t << 13; if(t) vl ^= ml;
							SHR1(bl, bm, bh, bt); SHR1(vl, vm, vh, vt);
						}
					} 
					while
						(at < bt || at == bt && (ah < bh || ah == bh && 
											(am < bm || am == bm && al < bl)));

					if(al == bl && am == bm && ah == bh && at == bt) 
						
						break;

					do 
					{
						al ^= bl; am ^= bm; ah ^= bh; at ^= bt;
						ul ^= vl; um ^= vm; uh ^= vh; ut ^= vt;

						sh = shiftTab[al & 63]; th = 32 - sh;
						t = ul & 63; ut ^= (u32)xorTab[t] << 13; ul ^= xorTab2[t];
						SHR(al, am, ah, at, sh, th); SHR(ul, um, uh, ut, sh, th);
            
						while(!(al & 1)) 
						{
							t = ul & 1; ut ^= t << 13; if(t) ul ^= ml;
							SHR1(al, am, ah, at); SHR1(ul, um, uh, ut);
						}
					} 
					while
						(at > bt || at == bt && (ah > bh || ah == bh && 
												(am > bm || am == bm && al > bl)));
				} 
				while(al != bl);
			} 
			while(am != bm);
		} 
		while(ah != bh);
	} 
	while(at != bt);

	// Now a (and b) must equal 1

	// Reduce u modulo t^109+t^9+t^2+t+1

	t = ut >> 13; ut ^= t << 13; ul ^= t; ul ^= t << 1; ul ^= t << 2; ul ^= t << 9;

    res->_3 = ut; res->_2 = uh; res->_1 = um; res->_0 = ul;
}

// Divide x by y in the field GF(2^109) i.e., as polys in (Z/2Z)[t]
// reduced modulo t^109+t^9+t^2+t+1.  Degree x, y < 109, y != 0

static void quotient(poly128 *res, const poly128 *x, const poly128 *y) 
{
	inverse(res, y); product(res, x);
}

// Return product x*y modulo m, 0 <= x, y < m <= 2^127
// Could this possibly be any slower?  =:-)

static void productMod(u128 *x, const u128 *y, const u128 *m) 
{	uint		i;
	u32			b, w, v[3];
	u128		z;

	v[0] = x->_0; v[1] = x->_1; v[2] = x->_2; w = x->_3;
  
	for(i = 3; !w && i; ) 
		
		w = v[--i];
  
	if(!w) return;

	b = top_bit(w); 
	
	*x = *y; z = *y;	// in case x = y
  
	while(b >>= 1) 
	{
		doubleMod(x, m);
    
		if(w & b) 
			
			sumMod(x, &z, m);
	}

	while(i) 
	{
		b = 1ul << 31; w = v[--i];
    
		do 
		{	doubleMod(x, m);
      
			if(w & b) 
				
				sumMod(x, &z, m);
      
			b >>= 1;
		} 
		while
			(b);
	}
}

// Return power x^y modulo m, 0 <= x < m <= 2^127, 0 <= y < 2^64.
// Note that 0^0 returns 1 (even when m = 1)

static void powerMod(u128 *r, const u128 *x, const u64 *y, const u128 *m) 
{	u32			b, w;

	if(!(w = y->_1) && !(w = y->_0))
	{
		r->_0 = 1; r->_1 = r->_2 = r->_3 = 0;
	}
	else
	{
		b = top_bit(w); *r = *x;
  
		while(b >>= 1) 
		{
			productMod(r, r, m);
    
			if(w & b) productMod(r, x, m);
		}

		if(y->_1) 
		{
			w = y->_0; b = 1ul << 31;
    
			do 
			{
				productMod(r, r, m);
      
				if(w & b) productMod(r, x, m);
			} 
			while(b >>= 1);
		}
	}
}

// Computes multiplier Prod_i pow(multipliers[i], cases[i]) modulo order

// Here multiplier[i] = (Frobenius to power i) + 1 modulo group order

static const u128 multipliers[CASES] = { 
	data_u128(0x6D3, 0x25F3CBE9, 0x20245E26, 0x59822D94),
	data_u128(0x6D3, 0x25F3CBE9, 0x20245E26, 0x59822D96),
	data_u128(0xB86, 0x8E249C44, 0x9EDF28ED, 0x4BE9F47A),
	data_u128(0x92C, 0xDA0C3416, 0xDF81C389, 0xD2B6110D),
	data_u128(0x21F, 0xBDC2FB8D, 0xA169935F, 0x671A66B6),
	data_u128(0xFC6, 0x09AA935F, 0xE1B24FAC, 0x1A1EC1D4),
	data_u128(0xB86, 0x8E249C44, 0x9EDF28ED, 0x4BE9F46A),
};

static void findMultiplier(u128 *r, u64 *cases) 
{	uint	i;
	u128	ff;

	powerMod(r, multipliers, cases, &order); cases->_1 = cases->_0 = 0;

	for(i = 1; i < CASES; ++i) 
	{
		powerMod(&ff, multipliers + i, cases + i, &order); 
		
		productMod(r, &ff, &order); (cases + i)->_1 = (cases + i)->_0 = 0;
	}
}

// ARITHMETIC ON ELLIPTIC CURVE OVER THE FIELD

// Given a point (x:y:z) on y^2 + x*y = x^3 + a*x^2 + b, this computes
// its double (x2:y2:z2) by the group law
// It puts x2 in *px2, y2 in *py2 and returns z2
// The point at infinity is represented by (0:1:0)
// Finite points are represented by (x:y:1)

static int ellipticDouble(const poly128 *x, const poly128 *y, int z, poly128 *px2, poly128 *py2) 
{	poly128			lam, x2,y2;

	if(!z || equal(x, &p128_0)) 
	{ 
		*px2 = p128_0; *py2 = p128_1; return 0; 
	}

	quotient(&lam, y, x); xor(&lam, x);
  
	squareNTimes(&x2, &lam, 1); xor(&x2, &lam); xor(&x2, &p128_a);

	y2 = *x; xor(&y2, &x2); product(&y2, &lam); xor(&y2, &x2); xor(&y2, y);

	*px2 = x2; *py2 = y2; return 1;
} 

// Given points (x1:y1:z1) and (x2:y2:z2) on y^2 + x*y = x^3 + a*x^2 + b,
// this computes their sum (x3:y3:z3) by the group law.  It puts x3 in 
// *px3, y3 in *py3 and returns z3.  The point at infinity is represented 
// by (0:1:0).  Finite points are represented by (x:y:1)

static int ellipticSum(const poly128 *x1, const poly128 *y1, int z1, 
					   const poly128 *x2, const poly128 *y2, int z2, 
							 poly128 *px3, poly128 *py3) 
{	poly128 lam, x3, y3;

	if(!z1) 
	{ 
		*px3 = *x2; *py3 = *y2; return z2; 
	}
  
	if(!z2) 
	{ 
		*px3 = *x1; *py3 = *y1; return z1; 
	}

	if(equal(x1, x2)) 
	{
		if(equal(y1, y2)) 
		
			return ellipticDouble(x1, y1, z1, px3, py3);
    
		*px3 = p128_0; *py3 = p128_1; return 0;
	}

	y3 = *y1; xor(&y3, y2); x3 = *x1; xor(&x3, x2); quotient(&lam, &y3, &x3);
	squareNTimes(&x3, &lam, 1); xor(&x3, &lam); xor(&x3, x2); xor(&x3, &p128_a); 
	y3 = lam; product(&y3, &x3); xor(&x3, x1); xor(&y3, &x3); xor(&y3, y1);

	*px3 = x3; *py3 = y3; return 1;
}

// Given a point (x:y:z) on y^2 + x*y = x^3 + a*x^2 + b, this computes
// its fac-th multiple (x2:y2:z2) by the group law.  It puts x2 in *px2, 
// y2 in *py2 and returns z2.  The point at infinity is represented by 
// (0:1:0).  Finite points are represented by (x:y:1)

static int ellipticProduct(const poly128 *x, const poly128 *y, 
								int z, const u128 *fac, poly128 *px2, poly128 *py2) 
{	int		z2;
	u32		f, i, m, tab[4];
	poly128	x2,y2;

	// Get most significant 32-bit word of fac

	tab[0] = fac->_0; tab[1] = fac->_1; tab[2] = fac->_2; tab[3] = fac->_3;
	
	for(i = 3; !(f = tab[i]) && i; --i) 
		;

	// Quick exit if fac was 0
	if(!f) 
	{
		*px2 = p128_0; *py2 = p128_1; return 0;
	}

	// Run through bits of most significant word

	m = top_bit(f); x2 = *x; y2 = *y; z2 = z;
  
	while(m >>= 1) 
	{
		z2 = ellipticDouble(&x2, &y2, z2, &x2, &y2);
    
		if(f & m) 
			
			z2 = ellipticSum(x, y, z, &x2, &y2, z2, &x2, &y2);
	}

	// Run through bits of less significant words
  
	while(i--) 
	{
		f = tab[i]; m = 1u << 31;
    
		do 
		{
			z2 = ellipticDouble(&x2, &y2, z2, &x2, &y2);
      
			if(f & m) 
				
				z2 = ellipticSum(x, y, z, &x2, &y2, z2, &x2, &y2);
			
			m >>= 1;
		} 
		while(m);
	}

	*px2 = x2; *py2 = y2; return z2;
} 

// Check that *pu, *pv are in range, update them using cases[] and reset it
// Then check that x, y are in range and that [u]P+[v]Q = (x:y:1)
// Note: used before reporting a distinguished point or saving state

static int syncAndCheck(u64 *cases, u128 *pu, u128 *pv, poly128 *x, poly128 *y) 
{	int		zc, zu, zv;
	u128	f;
	poly128	xc,yc, xu,yu, xv,yv;

	const int zP = 1, zQ = 1;

	if(ge(pu, &order)) 
	  
		return 0;
  
	if(ge(pv, &order)) 
	  
		return 0;

	findMultiplier(&f, cases); productMod(pu, &f, &order); productMod(pv, &f, &order);

	if(x->_3 > 0x00001FFF || y->_3 > 0x00001FFF) 
	  
		return 0;

	zu = ellipticProduct(&xP, &yP, zP, pu, &xu, &yu);
	zv = ellipticProduct(&xQ, &yQ, zQ, pv, &xv, &yv);
	zc = ellipticSum(&xu, &yu, zu, &xv, &yv, zv, &xc, &yc);

	if(!equal(&xc, x) || !equal(&yc, y) || zc != 1) 
		
		return 0;

	return 1;
}

// FUNCTIONS FOR INPUT and OUTPUT

FILE	*dpt_fh;	// for the distinguished points file
FILE	*log_fh;	// for the log file

// Prevent threads interfering with each other by using
// critical sections to control access to shared files
// This is done in ReadSate, WriteState and find_points

void string2pfile(char *s)		// write to distinguished points file
{
	fputs(s, dpt_fh);
}

void string2log(const char *s)	// write to log file
{
	fputs(s, log_fh); 
}

void int2log(int no)
{	char	buf[32];

	sprintf(buf, "%i", no); string2log(buf);
}

void uint2log(uint no)
{	char	buf[32];

	sprintf(buf, "%u", no); string2log(buf);
}

void double2log(double no)
{	char	buf[32];

	sprintf(buf, "%g", no); string2log(buf);
}

void err2log(void)
{
	string2log("\n         (errno = "); int2log(errno);

	string2log(strerror(errno)); string2log(").\n");
}

static void u32ToBytes(u32 data, u8 *p) 
{
  p[0] = data; p[1] = data >> 8; p[2] = data >> 16; p[3] = data >> 24;

}

static u32 bytesToU32(u8 *p) 
{
	return p[0] | (u32)p[1] << 8 | (u32)p[2] << 16 | (u32)p[3] << 24;
}

// Write state out to file in a portable binary format

#define OUT128(offset, x)				\
  u32ToBytes(x._0, buf + offset    );	\
  u32ToBytes(x._1, buf + offset + 4);	\
  u32ToBytes(x._2, buf + offset + 8);	\
  u32ToBytes(x._3, buf + offset +12) 

static void writeState(char *fn, u64 *itersT, u128 *uT, u128 *vT, poly128 *xT, poly128 *yT) 
{	FILE	*handle;
	uint	i;
	int		status;
	u8		buf[72];
	time_t	now;

	EnterCriticalSection(&crit);

	if(handle = fopen(fn, "wb"))
	{
		now = time(NULL); 
		
		string2log("Saving state to "); string2log(fn); 
		string2log(" on "); string2log(ctime(&now));

		for(i = 0; i < PARAL; ++i) 
		{
			u32ToBytes(itersT[i]._0, buf); u32ToBytes(itersT[i]._1, buf + 4);
			OUT128( 8, uT[i]); OUT128(24, vT[i]);
			OUT128(40, xT[i]); OUT128(56, yT[i]);

			if(fwrite((void *)buf, sizeof(buf), 1, handle) != 1) 
			{
				string2log("Warning: couldn't fwrite() to saved state file");

				err2log(); break;
			}
		}

		if((status = fclose(handle)) == EOF) 
		{
			string2log("Warning: fclose() of saved state file failed"); err2log();
		}

	} 
	else 
	{	
		string2log("Warning: couldn't fopen() saved state file for writing"); err2log();
	}

	fflush(log_fh); LeaveCriticalSection(&crit);
}

#define IN128(offset, x)				\
  x._0 = bytesToU32(buf + offset     );	\
  x._1 = bytesToU32(buf + offset +  4); \
  x._2 = bytesToU32(buf + offset +  8); \
  x._3 = bytesToU32(buf + offset + 12)

// Reads state back from file (if it exists) in a portable binary format
// and returns number of valid items (i.e., u,v,x,y,z data) found
// Note: There is nothing to check for the itersT[] array

static uint readState(char *fn, u64 *itersT, u128 *uT, u128 *vT, poly128 *xT, poly128 *yT) 
{	uint	i = 0;
	int		ok;
	FILE	*handle;

	EnterCriticalSection(&crit);

	if(handle = fopen(fn, "rb"))
	{	int		status;
		u8		buf[72];

		for(i = 0; i < PARAL; ++i) 
		{
			if(fread((void *)buf, sizeof(buf), 1, handle) != 1) 
			{
				if(feof(handle)) 
					
					string2log("Warning: short saved state file.\n");
        
				else 
				{
					string2log("Warning: couldn't fread() from saved state file"); 
					
					err2log(); break;
				}
			}

			itersT[i]._0 = bytesToU32(buf); itersT[i]._1 = bytesToU32(buf+4);
			IN128( 8, uT[i]); IN128(24, vT[i]);
			IN128(40, xT[i]); IN128(56, yT[i]);

			do
			{	int		z, zu, zv;
				poly128 x,y, xu,yu, xv,yv;

				const int zP = 1, zQ = 1;

				if(ge(uT + i, &order) || ge(vT + i, &order)) 
					
					break;

				if(xT[i]._3 > 0x00001FFF || yT[i]._3 > 0x00001FFF) 
					
					break;

				zu = ellipticProduct(&xP, &yP, zP, uT + i, &xu, &yu);
				zv = ellipticProduct(&xQ, &yQ, zQ, vT + i, &xv, &yv);
				z = ellipticSum(&xu, &yu, zu, &xv, &yv, zv, &x, &y);

				if(!equal(&x, xT + i) || !equal(&y, yT + i) || z != 1) 
					
					break;
        
				ok = 1;
			}
			while(0);

			if(!ok) 
			{
				string2log("Warning: invalid starting point found in saved state!\n");
        
				break;
			}
		}

		string2log("Read "); uint2log(i); string2log(" of "); 
		uint2log((uint)PARAL); string2log(" starting points from saved state.\n");

		if((status = fclose(handle)) == EOF) 
		{
			string2log("Warning: fclose() of saved state file failed"); err2log();
		}
	} 
	else 
	{		
		string2log("Warning: couldn't fopen() saved state file for reading\n"); err2log(); 
		string2log("Note: \'No such file\' is normal for the first run.\n");
	}

	fflush(log_fh); LeaveCriticalSection(&crit);

	return i;
}

// Get process (thread times)

double get_time(void)	// process time for thread
{	OsInfo	op;

	if(op.id() == OsInfo::WinNT)
	{	FILETIME	pct, pxt, pkt, put;

		GetThreadTimes(GetCurrentThread(), &pct, &pxt, &pkt, &put);

//		double kt = (double)pkt.dwHighDateTime * 4294967296.0 + pkt.dwLowDateTime;
		double ut = (double)put.dwHighDateTime * 4294967296.0 + put.dwLowDateTime;

//		return 1.0e-7 * (kt + ut);	// add user and kernel time for thread
		return 1.0e-7 * ut;	// only user time (as for Xavier's Cygwin port)
	}
	else if(op.id() == OsInfo::Win95 || op.id() == OsInfo::Win98)
	
		return (double)time(0);
	
	else
	
		return 0;
}

// CODE FOR SENDING RESULTS USING HTTP

#define	HTTP_URL			"http://cristal.inria.fr/bin/ecdl7"
#define	HTTP_SERVER_HOST	"cristal.inria.fr"

static struct sockaddr_in	http_addr;
static char				   *http_host = HTTP_SERVER_HOST;
static int					http_port = 80;
static char				   *da_host   = NULL;

static int http_parse_proxy(char *proxy)
{	char	*host;
	int		port;

	if((host = (char*)malloc(strlen(proxy))) == NULL)
	{		
		string2log("Error: cannot allocate proxy host name\n"); return -1;
	}

	if(sscanf(proxy, "%[^:]:%d", host, &port) != 2) 
	{
		string2log("Error: bad proxy specification `"); 
		string2log(proxy); string2log("'\n");
		
		free(host); return -1;
	}

	da_host = http_host = host; http_port = port; return 0;
}

static SOCKET http_open(void)
{	WORD			wVersionRequested;
	WSADATA			wsaData;
	struct hostent	*entry;
 
	wVersionRequested = MAKEWORD(1, 1);
 
	if(WSAStartup(wVersionRequested, &wsaData))
	{
		string2log("Warning: could not find network service\n"); return - 1;
	}

	if((entry = gethostbyname(http_host)) == NULL)
	{				
		string2log("Warning: \'"); string2log(http_host); string2log("\' could not be found\n");
		
		WSACleanup(); return -1; 
	}

	if(entry->h_length < 0 || (size_t)entry->h_length > sizeof(http_addr.sin_addr.s_addr)) 
	{
		string2log("Error: address buffer overflow!\n"); WSACleanup(); return -1; 
	}

	memset(&http_addr, 0, sizeof(struct sockaddr_in));
	http_addr.sin_family = AF_INET;
	memcpy(&http_addr.sin_addr.s_addr, entry->h_addr, entry->h_length);
	http_addr.sin_port = htons(http_port);

	return 0;
}

static void http_close(void)
{
	WSACleanup(); 

	if(da_host)

		free(da_host);
}

static int put_socket(SOCKET s, char *buf)
{	int		retcode, len;

	len = strlen(buf);

	retcode = send(s, buf, len, 0);

	if(retcode != len)

		return SOCKET_ERROR;

	return retcode;
}

static int http_error(char* host, char *err)
{	int	retcode;

	retcode = WSAGetLastError();
	
	if(*host == '\0')
	
		string2log("HTTP Error: "); 
		
	else
	{
		string2log("HTTP Error for host \'"); string2log(host); string2log("\': ");
	}

	string2log(err); string2log("(error code "); int2log(retcode); string2log(")\n");

	return -1;
}

// Send a string back to the base using the HTTP POST protocol

static int httpPost(char *payload)
{	int		retcode, bad;
	char	buf[256];
	SOCKET	s;

	if((s = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
	{				
		string2log("Warning: could not start an http request\n"); return -1; 
	}

	if(connect(s, (struct sockaddr*)&http_addr, sizeof(http_addr)))

		return http_error(http_host, "could not connect");

	sprintf(buf, "POST %s HTTP/1.0\r\n", HTTP_URL);  bad = 1;
		
	if(put_socket(s, buf) != SOCKET_ERROR)
	{
		sprintf(buf, "Host: %s:%d \r\n", http_host, http_port);
		
		if(put_socket(s, buf) != SOCKET_ERROR)
		{
			sprintf(buf, "Content-length: %lu\r\n\r\n", (unsigned long)strlen(payload));
		
			if(put_socket(s, buf) != SOCKET_ERROR)

				bad = (put_socket(s, payload) == SOCKET_ERROR);
		}
	}

	if(shutdown(s, 1) == SOCKET_ERROR || bad)

		return http_error(http_host, "could not send data");
  
	// Read response from server or proxy
  
	// RFC 2616: first line of response is the status line, with format
    // HTTP-Version SP Status-Code SP Reason-Phrase CRLF
	// But watch out for 1xx (informational) responses
  
	retcode = recv(s, buf, 255, 0);

	while(retcode && retcode != SOCKET_ERROR)
	{
		if(retcode < 256)

			buf[retcode] = '\0';

		if(sscanf(buf, "HTTP%*s %d ", &retcode) == 1) 
		{
			if(retcode >= 200) 
				
				break;
		}

		retcode = recv(s, buf, 255, 0);
	}

	// Check status code -- should be 2xx

	if(retcode < 200 || retcode >= 300)
	{
		string2log("Web server or proxy "); string2log(http_host); 
		string2log(" returned an error (status code = "); 
		int2log(retcode); string2log(")\n");
	}

	if(shutdown(s, 0) == SOCKET_ERROR)

		return http_error(http_host, "server error reported");
  
	if(closesocket(s) == SOCKET_ERROR)
	
		string2log("Warning: could not close an http request\n"); 

	return 0;
}

// REPORTING OF DISTINGUISHED POINTS BACK TO HEADQUARTERS

static void reportDistinguished(ModeType mode, char *fn, int p_no, 
						u64 *iters, u128 *u, u128 *v, u64 *total, double dStart) 
{
	const char format[] =
		"ECC-L0|" VARIANT "|" VERSION "|\n"
		"ECC-L1|%03X%08X%08X%08X|%03X%08X%08X%08X|\n"
		"ECC-L2|%s|\n"
		"ECC-L3|%04X%08X|\n"
		"ECC-L4|%s|\n",
	test_format[] =
		"TEST-L0|" VARIANT "|" VERSION "|\n"
		"TEST-L1|%03X%08X%08X%08X|%03X%08X%08X%08X|\n"
		"TEST-L2|%s|\n"
		"TEST-L3|%04X%08X|\n"
		"TEST-L4|%s|\n";

	int		status;
	time_t	now;

	// Output the distinguished point to stdout and the points file

	EnterCriticalSection(&crit);

	if(mode == Test) 
		
		dpt_fh = NULL;
	
	else if(!(dpt_fh = fopen(fn, "a")))
	{
		string2log("Warning: couldn't fopen() file "); string2log(fn);
		string2log(" for appending"); err2log();
		string2log("Following point is not saved!\n");
	}

	// Time and date

	now = time(NULL);
  
	if(now == -1) 
		
		string2log("Warning: can't determine time.\n");
  
	else 
	{	char *str1, *str2;

		str1 = mode == Test ? "\nTest" : "\nDistinguished"; 
		str2 = ctime(&now);
    
		string2log(str1); string2log(" point found on "); string2log(str2);
    
		if(dpt_fh) 
		{
			string2pfile(str1); 
			string2pfile(" point found on "); 
			string2pfile(str2);
		}
	} 

	// The distinguished point

	int	tlen = strlen(test_format), flen = strlen(format);
	flen = (flen > tlen ? flen : tlen);
	char *msg = (char*)malloc(flen	// size of fixed text and format strings
		+ 2 * (27 - 16)				// two 27-digit hex numbers on line 1
        + strlen(argv3) - 2			// the user ID on line 2
        + (12 - 8)					// one 12-digit hex number on line 3
        + strlen(argv5) - 2			// the machine ID on line 4
        + 1);						// the terminating '\0'

	if(msg == NULL) 
		
		string2log("Warning: couldn't allocate temp buffer for writing point.\n");
	
	else
	{
		sprintf(msg, mode == Test ? test_format : format, u->_3, u->_2, u->_1, u->_0, 
			v->_3, v->_2, v->_1, v->_0, argv3, iters->_1, iters->_0, argv5);
  
		string2log(msg);

		if(dpt_fh)	// Note: mode != Test
			
			string2pfile(msg);

		if(dpt_fh && (status = fclose(dpt_fh)) == EOF)
		{
			string2log("Warning: fclose() of file "); string2log(fn);
			string2log(" failed"); err2log();
		}

		// Output extra verbosity to stdout

		{	double dIters, dTotal, dNow, dUserTime; // For timing

			dNow = get_time();

			// Iteration counts

			dIters = (double)iters->_1*4294967296.0+(double)iters->_0;
			dTotal = (double)total->_1*4294967296.0+(double)total->_0;

			string2log("Iterations used = "); double2log(dIters);
			string2log(".\nTotal iterations = "); double2log(dTotal);
			string2log(".\n");

			// Iteration rate

			dUserTime = dNow - dStart;

			if(dUserTime) 
			{
				string2log("Rate = "); double2log(dTotal / dUserTime); 
				string2log(" per second.\n");
			}
		} 

		if(mode == Http) 
		{
			if(httpPost(msg) == -1)
        
				string2log("Warning: an error occurred while sending point via HTTP, send by hand!\n");
		}  
	
		free(msg);
	}

	fflush(log_fh); LeaveCriticalSection(&crit);
}

// PARSE COMMEAND LINE 

static ModeType parse(int argc, char *argv[]) 
{	char		ch, *str;

	if(argc != 6 && argc != 8) 
	  
		return Bad;

	str = argv[1];

	if(!strcmp(str, "test")) 
		
		mode = Test;

	else if(!strcmp(str, "batch")) 
	  
	  mode = Batch;
	
	else if(!strcmp(str, "mail")) 
		
		mode = Mail;
  
	else if(!strcmp(str, "alt")) 
		
		mode = Alt;
  
	else if(!strcmp(str, "http")) 
		
		mode = Http;
  
	else 
		
		return Bad;

	if(strcmp(argv[2], "by") ) 
		
		return Bad;

	for(str = argv[3]; (ch = *str) != '\0'; ++str) 
		
		if(ch == '|') 
			
			return Bad;

	if(strcmp(argv[4], "on")) 
		
		return Bad;

	for(str = argv[5]; (ch = *str) != '\0'; ++str) 
		
		if(ch == '|') 
			
			return Bad;

	if(argc == 8) 
	{
		if(strcmp(argv[6], "proxy")) 
			
			return Bad;
    
		if(http_parse_proxy(argv[7]) == -1) 
			
			return Bad;
	}

	argv3 = argv[3]; argv5 = argv[5]; 

	return mode;
}

// Signal catcher for SIGINT or SIGTERM
// Allows program to stop cleanly and save state
// User can send the signal by Ctrl-C or "kill <PID>", for instance

static void catchSignal(int sigNum) 
{
	gotSignal = 1;
}

// MAIN LOOP FOR FINDING POINTS

// This subroutine runs in multiple threads

static void find_points(void *p_no)
{
	const int		zP = 1, zQ = 1;

	uint			distPopc;
	u64				total;
	double			dStart;

	uint			pops[PARAL], i, j;
	u64				casesT[PARAL][CASES], itersT[PARAL];
	u128			uT[PARAL], vT[PARAL], seed;
	poly128			denT[PARAL], x1T[PARAL], x2T[PARAL], y1T[PARAL], y2T[PARAL];

	char			state_fn[_MAX_PATH];

	strcpy(state_fn, st_fpath); state_fn[strlen(state_fn) - 5] += (char)p_no; 

	// Read starting points from the saved state or generate new random ones

	i = (mode == Test ? 0 : readState(state_fn, itersT, uT, vT, x1T, y1T));

    if(i < PARAL) 
	{
		EnterCriticalSection(&crit);
		string2log("Generating "); uint2log((uint)PARAL - i); 
		string2log(" new starting points.\n");
		LeaveCriticalSection(&crit);

		if(mode == Test)	// Fixed seed for testing
		{
			seed._3 = 0; seed._2 = 0; seed._1 = 0x01234567; seed._0 = 0x89ABCDEF;
		} 
		else 
		{
			seed._3 = lrand(); seed._2 = lrand(); seed._1 = lrand(); seed._0 = lrand(); 
			seed._3 &= 0x00000FFF; seed._2 &= 0xFFFFFFFE;	// Ensure seed < order
		}

		for(; i < PARAL; ++i) // Generate random points
		{	int			z, zu, zv;
			poly128		xu, xv, yu, yv;	
			// Euler gamma times 2^108
			const u128	gamma = data_u128(0x0000093C, 0x467E37DB, 0xC7A4D1B, 0xE3F81015);

			itersT[i]._1 = itersT[i]._0 = 0;
        
			do			// Not terribly random, but it will do!
			{	
          		uT[i] = seed;
				zu = ellipticProduct(&xP, &yP, zP, &seed, &xu, &yu);
				sumMod(&seed, &gamma, &order);

				vT[i] = seed;
				zv = ellipticProduct(&xQ, &yQ, zQ, &seed, &xv, &yv);
				sumMod(&seed, &gamma, &order);

				z = ellipticSum(&xu, &yu, zu, &xv, &yv, zv, x1T + i, y1T + i);
			}
			while(!z);
		}
	}

    // Initialise counts of cases to 0

    for(i = 0; i < PARAL; ++i) 		
      for(j = 0; j < CASES; ++j) 
		  
		  casesT[i][j]._1 = casesT[i][j]._0 = 0;

	EnterCriticalSection(&crit);
	string2log("Computing iterations...\n"); fflush(log_fh);
	LeaveCriticalSection(&crit);

	distPopc = mode == Test ? TEST_POP_COUNT : DIST_POP_COUNT;

	total._1 = total._0 = 0; dStart = get_time();

	while(!gotSignal) 
	{
		int someWereDistinguished = 0;

		for(i = 0; i < PARAL; ++i)		// Check for distinguished points
		{	poly128		tt;

			while((change_base(&tt, x1T + i), pops[i] = popc(&tt)) == distPopc) 
			{	int z;
				// Point is distinguished  - check that the point is OK

				if(!syncAndCheck(casesT[i], uT + i, vT + i, x1T + i, y1T + i)) 
				{
					EnterCriticalSection(&crit);
					string2log("Fatal error: *** bad distinguished point detected! ***\n");
					LeaveCriticalSection(&crit);
					return;
				}

				reportDistinguished(mode, dp_fpath, (int)p_no, 
										itersT + i, uT + i, vT + i, &total, dStart);

				// Restart this point by incrementing v
        
				itersT[i]._1 = itersT[i]._0 = 0;

				do 
				{
					incMod(vT + i, &order);
					z = ellipticSum(x1T + i, y1T + i, 1, &xQ, &yQ, zQ, x1T + i, y1T + i);
				} 
				while(!z);

				someWereDistinguished = 1;	// Note this for use immediately below
			}
		}

		// Save state after a bunch of iterations or a distinguished point

		if((total._0 | total._1) && (total._0 & ((1ul << SAVED_STATE_SHIFT) - 1)) < PARAL
							|| someWereDistinguished) 
		{	
			for(i = 0; i < PARAL; ++i)		// Check all points
			{
				if(!syncAndCheck(casesT[i], uT + i, vT + i, x1T + i, y1T + i)) 
				{
					EnterCriticalSection(&crit);
					string2log("Fatal error: *** bad point detected! ***\n"); 
					LeaveCriticalSection(&crit);
					return;
				}
			}
  
			if(mode != Test)	// Save state (except when testing)
				
				writeState(state_fn, itersT, uT, vT, x1T, y1T);
		}

		// Get denominators for additions of points

		for(i = 0; i < PARAL; ++i) 
		{	uint	m;

			m = pops[i]; m -= 7 * (m * 9 >> 6);  // Note: m <= 109
			if(m >= 7) m -= 7; // Reduce modulo 7
      
			++casesT[i][m]._0; if(!casesT[i][m]._0) ++casesT[i][m]._1;

			m += 2; if(m == 3) m = 1;
			
			squareNTimes(x2T + i, x1T + i, m);
			denT[i] = x1T[i]; xor(denT + i, x2T + i);
			squareNTimes(y2T + i, y1T + i, m);
		}

		// Invert PARAL denominators with 1 inversion and 3*PARAL-3 mults

		{	poly128 ix, prod, q, prodT[PARAL];

			prod = denT[0];
      
			for(i = 1; i < PARAL; ++i) 
			{
				prodT[i] = prod; product(&prod, denT + i);
			}

			inverse(&q, &prod);

			for(i = PARAL; --i; ) 
			{
				ix = q; product(&ix, prodT + i); product(&q, denT + i); denT[i] = ix;
			}

			denT[0] = q;
		}

		// Get new points

		for(i = 0; i < PARAL; ++i)	// Always in general case
		{ 
			poly128 y1 = y1T[i], lam, t; 
		  
			lam = y1; xor(&lam, y2T + i); product(&lam, denT + i);
			squareNTimes(&t, &lam, 1); xor(&t, &lam); xor(&t, x2T + i);

			t._0 ^= p128_a._0; // t = xor(t, a) with a = 0 or 1
			xor(x1T + i, &t);
			y1T[i] = t; product(y1T + i, &lam); xor(y1T + i, x1T + i); xor(y1T + i, &y1);
      
			++itersT[i]._0; if(!itersT[i]._0) ++itersT[i]._1;
		}

	    total._0 += PARAL; total._1 += total._0 < PARAL;
	}

	EnterCriticalSection(&crit);
	string2log("Received signal.\n");
	LeaveCriticalSection(&crit);

	// Save state after receiving signal (except when testing)

    // Check all points
 
	for(i = 0; i < PARAL; ++i) 
	{
		if(!syncAndCheck(casesT[i], uT + i, vT + i, x1T + i, y1T + i)) 
		{
			EnterCriticalSection(&crit);
			string2log("Fatal error: *** bad point detected after signal! ***\n"); 
			LeaveCriticalSection(&crit);
			return;
		}
    }

	// Save state after receiving signal (except when testing)

	if(mode != Test) 
	
		writeState(state_fn, itersT, uT, vT, x1T, y1T);

	thr_active &= ~(1 << (unsigned long)p_no);	// signal that thread is ending
												// to main program
	return;
}

// MAIN PROGRAM - IMPLEMENTS MULTIPROCESSING OPERATION ON WINDOWS NT

static int prty[3] = { IDLE_PRIORITY_CLASS, NORMAL_PRIORITY_CLASS, HIGH_PRIORITY_CLASS };
static int relp[5] = { THREAD_PRIORITY_LOWEST, THREAD_PRIORITY_BELOW_NORMAL, 
					   THREAD_PRIORITY_NORMAL, THREAD_PRIORITY_ABOVE_NORMAL, 
					   THREAD_PRIORITY_HIGHEST };

int ecdl_start(int argc, char *argv[], FILE* flog, char *exe_path, int pr) 
{	unsigned long	thr_hnd[32], p_no, p_mask, pr_mask, sy_mask, pos;
	char			*cp;
	OsInfo			os;
	time_t			now;

	if(!thr_active)		// initialise the application
	{
		// set up paths for state save and distributed points files
		cp = strrchr(exe_path, '\\'); pos = (cp == NULL ? 0 : cp - exe_path + 1);
		strcpy(st_fpath, exe_path); strcpy(st_fpath + pos, SAVED_STATE_FILENAME);
		strcpy(dp_fpath, exe_path); strcpy(dp_fpath + pos, DIST_POINTS_FILENAME);

		if(sizeof(u32) != 4) 
		{
			string2log("Fatal error: *** the size of u32 is wrong.  Stopping. ***\n");
    
			return 0;
		} 

		init_program(); log_fh = flog;

		// Parse command line

		parse(argc, argv);
  
		if(mode == Bad) 
		{
			string2log("Syntax: "); string2log(argv[0]); 
			string2log(" <mode> by <email> on <machine>\n"
					"See top of source code or readMe file for details.  Stopping.\n");
			return 0;
		}

		// Initialize HTTP stuff if http mode selected

		if(mode == Http) 
		{
			if(http_open() == -1) 
			
				return 0;
		}

		// Put up banner

		now = time(NULL); 

		string2log("ECDL Service started on "); string2log(ctime(&now));
		string2log("This is Rob's ECDL program for "); string2log(os.name());
		string2log(", version " VERSION "\nMode ...... "); string2log(argv[1]);
		string2log("\nEmail ..... "); string2log(argv[3]);
		string2log("\nMachine ... "); string2log(argv[5]); string2log(".\n");

		// Initialisation

		signal(SIGINT, catchSignal); signal(SIGTERM, catchSignal);

		SetPriorityClass(GetCurrentProcess(), prty[pr / 10]);

		GetProcessAffinityMask(GetCurrentProcess(), &pr_mask, &sy_mask);

		InitializeCriticalSection(&crit);	// set up a critical section to 
											// control writing to dist.points
		p_no = 0; p_mask = 1;

 		while(p_mask)						// scan the processor mask
		{
			if(p_mask & sy_mask)			// if a processor is found start a thread
			{								// on it using our (idle) priority class
				thr_hnd[p_no] = _beginthread(find_points, 0, (void*)p_no);

				if(thr_hnd[p_no] == -1l)
				{
					string2log("Thread creation failed\n");
				}
				else
				{
					thr_active |= p_mask;	// set active thread mask
	
					SetThreadAffinityMask((HANDLE)thr_hnd[p_no], p_mask);

					SetThreadPriority((HANDLE)thr_hnd[p_no], relp[pr % 10]);
				}
			}

			p_mask <<= 1; p_no++;
		}
	}

	return 1;
}

void ecdl_stop(void)
{	time_t	now;

	gotSignal = 1;

	while(thr_active)	// wait until all threads terminate 
	{
		Sleep(100L);
	}

	if(mode == Http) 

			http_close();

	now = time(NULL); 

	DeleteCriticalSection(&crit);

	string2log("ECDL Service stopped on "); string2log(ctime(&now));
}
